"""Integration tests for crash database implementation for Launchpad."""

# Copyright (C) 2007 - 2009 Canonical Ltd.
# Authors: Martin Pitt <martin.pitt@ubuntu.com>
#          Markus Korn <thekorn@gmx.de>
#
# This program is free software; you can redistribute it and/or modify it
# under the terms of the GNU General Public License as published by the
# Free Software Foundation; either version 2 of the License, or (at your
# option) any later version.  See http://www.gnu.org/copyleft/gpl.html for
# the full text of the license.

# pylint: disable=too-many-lines
# TODO: Address following pylint complaints
# pylint: disable=invalid-name

import atexit
import email
import os
import shutil
import subprocess
import sys
import tempfile
import unittest
import unittest.mock
from unittest.mock import MagicMock

import pytest

try:
    from launchpadlib.errors import HTTPError

    IMPORT_ERROR = None
except ImportError as error:
    IMPORT_ERROR = error

from apport.crashdb_impl.launchpad import CrashDatabase
from apport.packaging_impl import impl as packaging
from apport.report import Report

_CACHE = {}


def cache(func):
    """Decorate a function/method to cache the result of its call.

    The cache is ignored if force_fresh is set to True.
    """

    def try_to_get_from_cache(*args, **kwargs):
        if kwargs.get("force_fresh", False):
            return func(*args, **kwargs)
        if func.__name__ not in _CACHE:
            _CACHE[func.__name__] = func(*args, **kwargs)
        return _CACHE[func.__name__]

    return try_to_get_from_cache


@pytest.mark.requires_internet
@unittest.skipIf(IMPORT_ERROR, f"Python module not available: {IMPORT_ERROR}")
@unittest.skipUnless(
    "TEST_LAUNCHPAD" in os.environ,
    "Need Launchpad access with bug control permission. Set TEST_LAUNCHPAD to run.",
)
class T(unittest.TestCase):
    # pylint: disable=protected-access,too-many-public-methods
    """Integration tests for apport.crashdb_impl.launchpad."""

    # this assumes that a source package 'coreutils' exists and builds a
    # binary package 'coreutils'
    test_package = "coreutils"
    test_srcpackage = "coreutils"

    #
    # Generic tests, should work for all CrashDB implementations
    #

    def setUp(self):
        self.crashdb = self._get_instance()

        # create a local reference report so that we can compare
        # DistroRelease, Architecture, etc.
        self.ref_report = Report()
        self.ref_report.add_os_info()
        self.ref_report.add_user_info()
        self.ref_report["SourcePackage"] = "coreutils"

        # Objects tests rely on.
        self._create_project("langpack-o-matic")

    def _create_bug_from_report(self, name: str, report: Report) -> int:
        """Create a Launchpad bug report from a crash report.

        Return the bug ID.
        """
        bug_target = self._get_bug_target(self.crashdb, report)
        self.assertTrue(bug_target)

        crash_id, web_link = self._file_bug(bug_target, report)
        self.assertTrue(crash_id > 0)

        sys.stderr.write(f"(Created {name} report: {web_link}) ")
        return crash_id

    def _create_project(self, name):
        """Create a project using launchpadlib to be used by tests."""
        project = self.crashdb.launchpad.projects[name]
        if not project:
            self.crashdb.launchpad.projects.new_project(
                description=f"{name}description",
                display_name=name,
                name=name,
                summary=f"{name}summary",
                title=f"{name}title",
            )

    @property
    def hostname(self):
        """Get the Launchpad hostname for the given crashdb."""
        return self.crashdb.get_hostname()

    @cache
    def get_segv_report(self, force_fresh=False):
        # force_fresh used by @cache, pylint: disable=unused-argument
        """Generate SEGV crash report.

        This is only done once, subsequent calls will return the already
        existing ID, unless force_fresh is True.

        Return the ID.
        """
        r = self._generate_sigsegv_report()

        return self._create_bug_from_report("SEGV", r)

    @cache
    def get_python_report(self):
        """Generate Python crash report.

        Return the ID.
        """
        r = Report("Crash")
        r["ExecutablePath"] = "/bin/foo"
        r[
            "Traceback"
        ] = """Traceback (most recent call last):
  File "/bin/foo", line 67, in fuzz
    print(weird)
NameError: global name 'weird' is not defined"""
        r["Tags"] = "boogus pybogus"
        r.add_package_info(self.test_package)
        r.add_os_info()
        r.add_user_info()
        self.assertEqual(
            r.standard_title(),
            "foo crashed with NameError in fuzz():"
            " global name 'weird' is not defined",
        )

        return self._create_bug_from_report("Python", r)

    @cache
    def get_uncommon_description_report(self, force_fresh=False):
        # force_fresh used by @cache, pylint: disable=unused-argument
        """File a bug report with an uncommon description.

        This is only done once, subsequent calls will return the already
        existing ID, unless force_fresh is True.

        Example taken from real LP bug 269539. It contains only
        ProblemType/Architecture/DistroRelease in the description, and has
        free-form description text after the Apport data.

        Return the ID.
        """
        desc = """problem

ProblemType: Package
Architecture: amd64
DistroRelease: Ubuntu 8.10

more text

and more
"""
        bug = self.crashdb.launchpad.bugs.createBug(
            title="mixed description bug",
            description=desc,
            target=self.crashdb.lp_distro,
        )
        sys.stderr.write(
            f"(Created uncommon description:"
            f" https://{self.hostname}/bugs/{bug.id}) "
        )

        return bug.id

    def test_1_download(self):
        """download()"""
        r = self.crashdb.download(self.get_segv_report())
        self.assertEqual(r["ProblemType"], "Crash")
        self.assertEqual(r["Title"], "crash crashed with SIGSEGV in f()")
        self.assertEqual(r["DistroRelease"], self.ref_report["DistroRelease"])
        self.assertEqual(r["Architecture"], self.ref_report["Architecture"])
        self.assertEqual(r["Uname"], self.ref_report["Uname"])
        self.assertEqual(
            r.get("NonfreeKernelModules"), self.ref_report.get("NonfreeKernelModules")
        )
        self.assertEqual(r.get("UserGroups"), self.ref_report.get("UserGroups"))
        self.assertEqual(
            r.get_tags(),
            set(
                [
                    self.crashdb.arch_tag,
                    "apport-crash",
                    packaging.get_system_architecture(),
                ]
            ),
        )

        self.assertEqual(r["Signal"], "11")
        self.assertTrue(r["ExecutablePath"].endswith("/crash"))
        self.assertEqual(r["SourcePackage"], self.test_srcpackage)
        self.assertTrue(r["Package"].startswith(f"{self.test_package} "))
        self.assertIn("f (x=42)", r["Stacktrace"])
        self.assertIn("f (x=42)", r["StacktraceTop"])
        self.assertIn("f (x=42)", r["ThreadStacktrace"])
        self.assertGreater(len(r["CoreDump"]), 1000)
        self.assertIn("Dependencies", r)
        self.assertIn("Disassembly", r)
        self.assertIn("Registers", r)

        # check tags
        r = self.crashdb.download(self.get_python_report())
        self.assertEqual(
            r.get_tags(),
            set(
                [
                    "apport-crash",
                    "boogus",
                    "pybogus",
                    "need-duplicate-check",
                    packaging.get_system_architecture(),
                ]
            ),
        )

    def test_2_update_traces(self):
        # TODO: Split into separate test cases
        # pylint: disable=too-many-statements
        """update_traces()"""
        r = self.crashdb.download(self.get_segv_report())
        self.assertIn("CoreDump", r)
        self.assertIn("Dependencies", r)
        self.assertIn("Disassembly", r)
        self.assertIn("Registers", r)
        self.assertIn("Stacktrace", r)
        self.assertIn("ThreadStacktrace", r)
        self.assertEqual(r["Title"], "crash crashed with SIGSEGV in f()")

        # updating with a useless stack trace retains core dump
        r["StacktraceTop"] = "?? ()"
        r["Stacktrace"] = "long\ntrace"
        r["ThreadStacktrace"] = "thread\neven longer\ntrace"
        r["FooBar"] = "bogus"
        self.crashdb.update_traces(
            self.get_segv_report(), r, "I can has a better retrace?"
        )
        r = self.crashdb.download(self.get_segv_report())
        self.assertIn("CoreDump", r)
        self.assertIn("Dependencies", r)
        self.assertIn("Disassembly", r)
        self.assertIn("Registers", r)
        self.assertIn("Stacktrace", r)  # TODO: ascertain that it's the updated one
        self.assertIn("ThreadStacktrace", r)
        self.assertNotIn("FooBar", r)
        self.assertEqual(r["Title"], "crash crashed with SIGSEGV in f()")

        tags = self.crashdb.launchpad.bugs[self.get_segv_report()].tags
        self.assertIn("apport-crash", tags)
        self.assertNotIn("apport-collected", tags)

        # updating with a useful stack trace removes core dump
        r["StacktraceTop"] = (
            "read () from /lib/libc.6.so\nfoo (i=1) from /usr/lib/libfoo.so"
        )
        r["Stacktrace"] = "long\ntrace"
        r["ThreadStacktrace"] = "thread\neven longer\ntrace"
        self.crashdb.update_traces(self.get_segv_report(), r, "good retrace!")
        r = self.crashdb.download(self.get_segv_report())
        self.assertNotIn("CoreDump", r)
        self.assertIn("Dependencies", r)
        self.assertIn("Disassembly", r)
        self.assertIn("Registers", r)
        self.assertIn("Stacktrace", r)
        self.assertIn("ThreadStacktrace", r)
        self.assertNotIn("FooBar", r)

        # as previous title had standard form, the top function gets
        # updated
        self.assertEqual(r["Title"], "crash crashed with SIGSEGV in read()")

        # respects title amendments
        bug = self.crashdb.launchpad.bugs[self.get_segv_report()]
        bug.title = "crash crashed with SIGSEGV in f() on exit"
        try:
            bug.lp_save()
        except HTTPError:
            pass  # LP#336866 workaround
        r["StacktraceTop"] = (
            "read () from /lib/libc.6.so\nfoo (i=1) from /usr/lib/libfoo.so"
        )
        self.crashdb.update_traces(
            self.get_segv_report(), r, "good retrace with title amendment"
        )
        r = self.crashdb.download(self.get_segv_report())
        self.assertEqual(r["Title"], "crash crashed with SIGSEGV in read() on exit")

        # does not destroy custom titles
        bug = self.crashdb.launchpad.bugs[self.get_segv_report()]
        bug.title = "crash is crashy"
        try:
            bug.lp_save()
        except HTTPError:
            pass  # LP#336866 workaround

        r["StacktraceTop"] = (
            "read () from /lib/libc.6.so\nfoo (i=1) from /usr/lib/libfoo.so"
        )
        self.crashdb.update_traces(
            self.get_segv_report(), r, "good retrace with custom title"
        )
        r = self.crashdb.download(self.get_segv_report())
        self.assertEqual(r["Title"], "crash is crashy")

        # test various situations which caused crashes
        r["Stacktrace"] = ""  # empty file
        r["ThreadStacktrace"] = '"]\xb6"\n'  # not interpretable as UTF-8, LP #353805
        r["StacktraceSource"] = "a\nb\nc\ne\nf\n\xff\xff\xff\n\f"
        self.crashdb.update_traces(self.get_segv_report(), r, "tests")

    def test_get_comment_url(self):
        """get_comment_url() for non-ASCII titles"""
        title = b"1\xc3\xa4\xe2\x99\xa52"

        # distro, UTF-8 bytestring
        r = Report("Bug")
        r["Title"] = title
        url = self.crashdb.get_comment_url(r, 42)
        self.assertTrue(
            url.endswith("/ubuntu/+filebug/42?field.title=1%C3%A4%E2%99%A52")
        )

        # distro, unicode
        r["Title"] = title.decode("UTF-8")
        url = self.crashdb.get_comment_url(r, 42)
        self.assertTrue(
            url.endswith("/ubuntu/+filebug/42?field.title=1%C3%A4%E2%99%A52")
        )

        # package, unicode
        r["SourcePackage"] = "coreutils"
        url = self.crashdb.get_comment_url(r, 42)
        self.assertTrue(
            url.endswith(
                "/ubuntu/+source/coreutils/+filebug/42?field.title=1%C3%A4%E2%99%A52"
            )
        )

    def test_update_description(self):
        """update() with changing description"""
        bug_target = self.crashdb.lp_distro.getSourcePackage(name="bash")
        bug = self.crashdb.launchpad.bugs.createBug(
            description="test description for test bug.",
            target=bug_target,
            title="testbug",
        )
        crash_id = bug.id
        self.assertTrue(crash_id > 0)
        sys.stderr.write(f"(https://{self.hostname}/bugs/{crash_id}) ")

        r = Report("Bug")

        r["OneLiner"] = b"bogus\xe2\x86\x92".decode("UTF-8")
        r["StacktraceTop"] = "f()\ng()\nh(1)"
        r["ShortGoo"] = "lineone\nlinetwo"
        r["DpkgTerminalLog"] = "one\ntwo\nthree\nfour\nfive\nsix"
        r["VarLogDistupgradeBinGoo"] = b"\x01" * 1024

        self.crashdb.update(crash_id, r, "NotMe", change_description=True)

        r = self.crashdb.download(crash_id)

        self.assertEqual(r["OneLiner"], b"bogus\xe2\x86\x92".decode("UTF-8"))
        self.assertEqual(r["ShortGoo"], "lineone\nlinetwo")
        self.assertEqual(r["DpkgTerminalLog"], "one\ntwo\nthree\nfour\nfive\nsix")
        self.assertEqual(r["VarLogDistupgradeBinGoo"], b"\x01" * 1024)

        self.assertEqual(
            self.crashdb.launchpad.bugs[crash_id].tags, ["apport-collected"]
        )

    def test_update_comment(self):
        """update() with appending comment"""
        bug_target = self.crashdb.lp_distro.getSourcePackage(name="bash")
        # we need to fake an apport description separator here, since we
        # want to be lazy and use download() for checking the result
        bug = self.crashdb.launchpad.bugs.createBug(
            description="Pr0blem\n\n--- \nProblemType: Bug",
            target=bug_target,
            title="testbug",
        )
        crash_id = bug.id
        self.assertTrue(crash_id > 0)
        sys.stderr.write(f"(https://{self.hostname}/bugs/{crash_id}) ")

        r = Report("Bug")

        r["OneLiner"] = "bogus→"
        r["StacktraceTop"] = "f()\ng()\nh(1)"
        r["ShortGoo"] = "lineone\nlinetwo"
        r["DpkgTerminalLog"] = "one\ntwo\nthree\nfour\nfive\nsix"
        r["VarLogDistupgradeBinGoo"] = "\x01" * 1024

        self.crashdb.update(crash_id, r, "meow", change_description=False)

        r = self.crashdb.download(crash_id)

        self.assertNotIn("OneLiner", r)
        self.assertNotIn("ShortGoo", r)
        self.assertEqual(r["ProblemType"], "Bug")
        self.assertEqual(r["DpkgTerminalLog"], "one\ntwo\nthree\nfour\nfive\nsix")
        self.assertEqual(r["VarLogDistupgradeBinGoo"], "\x01" * 1024)

        self.assertEqual(
            self.crashdb.launchpad.bugs[crash_id].tags, ["apport-collected"]
        )

    def test_update_filter(self):
        """update() with a key filter"""
        bug_target = self.crashdb.lp_distro.getSourcePackage(name="bash")
        bug = self.crashdb.launchpad.bugs.createBug(
            description="test description for test bug",
            target=bug_target,
            title="testbug",
        )
        crash_id = bug.id
        self.assertTrue(crash_id > 0)
        sys.stderr.write(f"(https://{self.hostname}/bugs/{crash_id}) ")

        r = Report("Bug")

        r["OneLiner"] = "bogus→"
        r["StacktraceTop"] = "f()\ng()\nh(1)"
        r["ShortGoo"] = "lineone\nlinetwo"
        r["DpkgTerminalLog"] = "one\ntwo\nthree\nfour\nfive\nsix"
        r["VarLogDistupgradeBinGoo"] = "\x01" * 1024

        self.crashdb.update(
            crash_id,
            r,
            "NotMe",
            change_description=True,
            key_filter=["ProblemType", "ShortGoo", "DpkgTerminalLog"],
        )

        r = self.crashdb.download(crash_id)

        self.assertNotIn("OneLiner", r)
        self.assertEqual(r["ShortGoo"], "lineone\nlinetwo")
        self.assertEqual(r["ProblemType"], "Bug")
        self.assertEqual(r["DpkgTerminalLog"], "one\ntwo\nthree\nfour\nfive\nsix")
        self.assertNotIn("VarLogDistupgradeBinGoo", r)

        self.assertEqual(self.crashdb.launchpad.bugs[crash_id].tags, [])

    def test_get_distro_release(self):
        """get_distro_release()"""
        self.assertEqual(
            self.crashdb.get_distro_release(self.get_segv_report()),
            self.ref_report["DistroRelease"],
        )

    def test_get_affected_packages(self):
        """get_affected_packages()"""
        self.assertEqual(
            self.crashdb.get_affected_packages(self.get_segv_report()),
            [self.ref_report["SourcePackage"]],
        )

    def test_is_reporter(self):
        """is_reporter()"""
        self.assertTrue(self.crashdb.is_reporter(self.get_segv_report()))
        self.assertFalse(self.crashdb.is_reporter(1))

    def test_can_update(self):
        """can_update()"""
        self.assertTrue(self.crashdb.can_update(self.get_segv_report()))
        self.assertFalse(self.crashdb.can_update(1))

    def test_duplicates(self):
        """Test duplicate handling."""
        # initially we have no dups
        self.assertIsNone(self.crashdb.duplicate_of(self.get_segv_report()))
        self.assertIsNone(self.crashdb.get_fixed_version(self.get_segv_report()))

        segv_id = self.get_segv_report()
        known_test_id = self.get_uncommon_description_report()
        known_test_id2 = self.get_uncommon_description_report(force_fresh=True)

        # dupe our segv_report and check that it worked; then undupe it
        r = self.crashdb.download(segv_id)
        self.crashdb.close_duplicate(r, segv_id, known_test_id)
        self.assertEqual(self.crashdb.duplicate_of(segv_id), known_test_id)

        # this should be a no-op
        self.crashdb.close_duplicate(r, segv_id, known_test_id)
        self.assertEqual(self.crashdb.duplicate_of(segv_id), known_test_id)

        self.assertEqual(self.crashdb.get_fixed_version(segv_id), "invalid")
        self.crashdb.close_duplicate(r, segv_id, None)
        self.assertIsNone(self.crashdb.duplicate_of(segv_id))
        self.assertIsNone(self.crashdb.get_fixed_version(segv_id))

        # this should have removed attachments; note that Stacktrace is
        # short, and thus inline
        r = self.crashdb.download(self.get_segv_report())
        self.assertNotIn("CoreDump", r)
        self.assertNotIn("Disassembly", r)
        self.assertNotIn("ProcMaps", r)
        self.assertNotIn("ProcStatus", r)
        self.assertNotIn("Registers", r)
        self.assertNotIn("ThreadStacktrace", r)

        # now try duplicating to a duplicate bug; this should automatically
        # transition to the master bug
        self.crashdb.close_duplicate(Report(), known_test_id, known_test_id2)
        self.crashdb.close_duplicate(r, segv_id, known_test_id)
        self.assertEqual(self.crashdb.duplicate_of(segv_id), known_test_id2)

        self.crashdb.close_duplicate(Report(), known_test_id, None)
        self.crashdb.close_duplicate(Report(), known_test_id2, None)
        self.crashdb.close_duplicate(r, segv_id, None)

        # this should be a no-op
        self.crashdb.close_duplicate(Report(), known_test_id, None)
        self.assertIsNone(self.crashdb.duplicate_of(known_test_id))

        self.crashdb.mark_regression(segv_id, known_test_id)
        self._verify_marked_regression(segv_id)

    def test_marking_segv(self):
        """Test processing status markings for signal crashes."""
        # mark_retraced()
        unretraced_before = self.crashdb.get_unretraced()
        self.assertIn(self.get_segv_report(), unretraced_before)
        self.assertNotIn(self.get_python_report(), unretraced_before)
        self.crashdb.mark_retraced(self.get_segv_report())
        unretraced_after = self.crashdb.get_unretraced()
        self.assertNotIn(self.get_segv_report(), unretraced_after)
        self.assertEqual(
            unretraced_before, unretraced_after.union(set([self.get_segv_report()]))
        )
        self.assertIsNone(self.crashdb.get_fixed_version(self.get_segv_report()))

        # mark_retrace_failed()
        self._mark_needs_retrace(self.get_segv_report())
        self.crashdb.mark_retraced(self.get_segv_report())
        self.crashdb.mark_retrace_failed(self.get_segv_report())
        unretraced_after = self.crashdb.get_unretraced()
        self.assertNotIn(self.get_segv_report(), unretraced_after)
        self.assertEqual(
            unretraced_before, unretraced_after.union(set([self.get_segv_report()]))
        )
        self.assertIsNone(self.crashdb.get_fixed_version(self.get_segv_report()))

        # mark_retrace_failed() of invalid bug
        self._mark_needs_retrace(self.get_segv_report())
        self.crashdb.mark_retraced(self.get_segv_report())
        self.crashdb.mark_retrace_failed(self.get_segv_report(), "I don't like you")
        unretraced_after = self.crashdb.get_unretraced()
        self.assertNotIn(self.get_segv_report(), unretraced_after)
        self.assertEqual(
            unretraced_before, unretraced_after.union(set([self.get_segv_report()]))
        )
        self.assertEqual(
            self.crashdb.get_fixed_version(self.get_segv_report()), "invalid"
        )

    def test_marking_project(self):
        """Test processing status markings for a project CrashDB."""
        # create a distro bug
        distro_bug = self.crashdb.launchpad.bugs.createBug(
            description="foo",
            tags=self.crashdb.arch_tag,
            target=self.crashdb.lp_distro,
            title="ubuntu distro retrace bug",
        )

        # create a project crash DB and a bug
        launchpad_instance = os.environ.get("APPORT_LAUNCHPAD_INSTANCE") or "qastaging"

        project_db = CrashDatabase(
            os.environ.get("LP_CREDENTIALS"),
            {"project": "langpack-o-matic", "launchpad_instance": launchpad_instance},
        )
        project_bug = project_db.launchpad.bugs.createBug(
            description="bar",
            tags=project_db.arch_tag,
            target=project_db.lp_distro,
            title="project retrace bug",
        )

        # on project_db, we recognize the project bug and can mark it
        unretraced_before = project_db.get_unretraced()
        self.assertIn(project_bug.id, unretraced_before)
        self.assertNotIn(distro_bug.id, unretraced_before)
        project_db.mark_retraced(project_bug.id)
        unretraced_after = project_db.get_unretraced()
        self.assertNotIn(project_bug.id, unretraced_after)
        self.assertEqual(
            unretraced_before, unretraced_after.union(set([project_bug.id]))
        )
        self.assertIsNone(self.crashdb.get_fixed_version(project_bug.id))

    def test_marking_foreign_arch(self):
        """Test processing status markings for a project CrashDB."""
        # create a DB for fake arch
        launchpad_instance = os.environ.get("APPORT_LAUNCHPAD_INSTANCE") or "qastaging"
        fakearch_db = CrashDatabase(
            os.environ.get("LP_CREDENTIALS"),
            {
                "distro": "ubuntu",
                "launchpad_instance": launchpad_instance,
                "architecture": "fakearch",
            },
        )

        fakearch_unretraced_before = fakearch_db.get_unretraced()
        systemarch_unretraced_before = self.crashdb.get_unretraced()

        # create a bug with a fake architecture
        bug = self.crashdb.launchpad.bugs.createBug(
            description="foo",
            tags=["need-fakearch-retrace"],
            target=self.crashdb.lp_distro,
            title="ubuntu distro retrace bug for fakearch",
        )
        print(f"fake arch bug: https://staging.launchpad.net/bugs/{bug.id}")

        fakearch_unretraced_after = fakearch_db.get_unretraced()
        systemarch_unretraced_after = self.crashdb.get_unretraced()

        self.assertEqual(systemarch_unretraced_before, systemarch_unretraced_after)
        self.assertEqual(
            fakearch_unretraced_after, fakearch_unretraced_before.union(set([bug.id]))
        )

    def test_marking_python(self):
        """Test processing status markings for interpreter crashes."""
        unchecked_before = self.crashdb.get_dup_unchecked()
        self.assertIn(self.get_python_report(), unchecked_before)
        self.assertNotIn(self.get_segv_report(), unchecked_before)
        self.crashdb._mark_dup_checked(self.get_python_report(), self.ref_report)
        unchecked_after = self.crashdb.get_dup_unchecked()
        self.assertNotIn(self.get_python_report(), unchecked_after)
        self.assertEqual(
            unchecked_before, unchecked_after.union(set([self.get_python_report()]))
        )
        self.assertIsNone(self.crashdb.get_fixed_version(self.get_python_report()))

    def test_update_traces_invalid(self):
        """Test updating an invalid crash.

        This simulates a race condition where a crash being processed gets
        invalidated by marking it as a duplicate.
        """
        crash_id = self.get_segv_report(force_fresh=True)

        r = self.crashdb.download(crash_id)

        self.crashdb.close_duplicate(r, crash_id, self.get_segv_report())

        # updating with a useful stack trace removes core dump
        r["StacktraceTop"] = (
            "read () from /lib/libc.6.so\nfoo (i=1) from /usr/lib/libfoo.so"
        )
        r["Stacktrace"] = "long\ntrace"
        r["ThreadStacktrace"] = "thread\neven longer\ntrace"
        self.crashdb.update_traces(crash_id, r, "good retrace!")

        r = self.crashdb.download(crash_id)
        self.assertNotIn("CoreDump", r)

    @unittest.mock.patch.object(CrashDatabase, "_get_source_version", MagicMock())
    def test_get_fixed_version(self):
        """get_fixed_version() for fixed bugs

        Other cases are already checked in test_marking_segv() (invalid
        bugs) and test_duplicates (duplicate bugs) for efficiency.
        """
        # staging.launchpad.net often does not have Quantal, so mock-patch
        # it to a known value
        CrashDatabase._get_source_version.return_value = "3.14"
        self._mark_report_fixed(self.get_segv_report())
        fixed_ver = self.crashdb.get_fixed_version(self.get_segv_report())
        self.assertEqual(fixed_ver, "3.14")
        self._mark_report_new(self.get_segv_report())
        self.assertIsNone(self.crashdb.get_fixed_version(self.get_segv_report()))

    #
    # Launchpad specific implementation and tests
    #

    @staticmethod
    @cache
    def _get_instance():
        """Create a CrashDB instance."""
        launchpad_instance = os.environ.get("APPORT_LAUNCHPAD_INSTANCE") or "qastaging"

        return CrashDatabase(
            os.environ.get("LP_CREDENTIALS"),
            {"distro": "ubuntu", "launchpad_instance": launchpad_instance},
        )

    @staticmethod
    def _get_bug_target(db, report):
        """Return the bug_target for this report."""
        project = db.options.get("project")
        if "SourcePackage" in report:
            return db.lp_distro.getSourcePackage(name=report["SourcePackage"])
        if project:
            return db.launchpad.projects[project]
        return None

    def _file_bug(
        self, bug_target: object, report: Report, description: str = "some description"
    ) -> tuple[int, str]:
        """File a bug report for a report.

        Return the bug ID.
        """
        # unfortunately staging's +storeblob API hardly ever works, so we
        # must avoid using it. Fake it by manually doing the comments and
        # attachments that +filebug would ordinarily do itself when given a
        # blob handle.
        mime = self.crashdb._generate_upload_blob(report)
        msg = email.message_from_binary_file(mime)
        mime.close()
        msg_iter = msg.walk()

        # first one is the multipart container
        header = next(msg_iter)
        assert header.is_multipart()

        # second part should be an inline text/plain attachments
        # with all short fields
        part = next(msg_iter)
        assert not part.is_multipart()
        assert part.get_content_type() == "text/plain"
        payload = part.get_payload(decode=True)
        assert isinstance(payload, bytes)
        description += "\n\n" + payload.decode("UTF-8", "replace")

        # create the bug from header and description data
        bug = self.crashdb.launchpad.bugs.createBug(
            description=description,
            # temporarily disabled to work around SSLHandshakeError on
            # private attachments
            # private=(header['Private'] == 'yes'),
            tags=header["Tags"].split(),
            target=bug_target,
            title=report.get("Title", report.standard_title()),
        )

        # now add the attachments
        for part in msg_iter:
            assert not part.is_multipart()
            bug.addAttachment(
                comment="",
                description=part.get_filename(),
                content_type=None,
                data=part.get_payload(decode=True),
                filename=part.get_filename(),
                is_patch=False,
            )

        for subscriber in header["Subscribers"].split():
            sub = self.crashdb.launchpad.people[subscriber]
            if sub:
                bug.subscribe(person=sub)

        return bug.id, bug.web_link

    def _mark_needs_retrace(self, crash_id):
        """Mark a report ID as needing retrace."""
        bug = self.crashdb.launchpad.bugs[crash_id]
        if self.crashdb.arch_tag not in bug.tags:
            bug.tags = bug.tags + [self.crashdb.arch_tag]
            bug.lp_save()

    def _mark_needs_dupcheck(self, crash_id):
        """Mark a report ID as needing duplicate check."""
        bug = self.crashdb.launchpad.bugs[crash_id]
        if "need-duplicate-check" not in bug.tags:
            bug.tags = bug.tags + ["need-duplicate-check"]
            bug.lp_save()

    def _mark_report_fixed(self, crash_id):
        """Close a report ID as "fixed"."""
        bug = self.crashdb.launchpad.bugs[crash_id]
        tasks = list(bug.bug_tasks)
        assert len(tasks) == 1
        t = tasks[0]
        t.status = "Fix Released"
        t.lp_save()

    def _mark_report_new(self, crash_id):
        """Reopen a report ID as "new"."""
        bug = self.crashdb.launchpad.bugs[crash_id]
        tasks = list(bug.bug_tasks)
        assert len(tasks) == 1
        t = tasks[0]
        t.status = "New"
        t.lp_save()

    def _verify_marked_regression(self, crash_id):
        """Verify that report ID is marked as regression."""
        bug = self.crashdb.launchpad.bugs[crash_id]
        self.assertIn("regression-retracer", bug.tags)

    def test_project(self):
        """Test reporting crashes against a project instead of a distro."""
        launchpad_instance = os.environ.get("APPORT_LAUNCHPAD_INSTANCE") or "qastaging"
        # crash database for langpack-o-matic project (this does not have
        # packages in any distro)
        crashdb = CrashDatabase(
            os.environ.get("LP_CREDENTIALS"),
            {"project": "langpack-o-matic", "launchpad_instance": launchpad_instance},
        )
        self.assertIsNone(crashdb.distro)

        # create Python crash report
        r = Report("Crash")
        r["ExecutablePath"] = "/bin/foo"
        r[
            "Traceback"
        ] = """Traceback (most recent call last):
  File "/bin/foo", line 67, in fuzz
    print(weird)
NameError: global name 'weird' is not defined"""
        r.add_os_info()
        r.add_user_info()
        self.assertEqual(
            r.standard_title(),
            "foo crashed with NameError in fuzz():"
            " global name 'weird' is not defined",
        )

        # file it
        bug_target = self._get_bug_target(crashdb, r)
        self.assertEqual(bug_target.name, "langpack-o-matic")

        crash_id, web_link = self._file_bug(bug_target, r)
        self.assertTrue(crash_id > 0)
        sys.stderr.write(f"({web_link}) ")

        # update
        r = crashdb.download(crash_id)
        r["StacktraceTop"] = (
            "read () from /lib/libc.6.so\nfoo (i=1) from /usr/lib/libfoo.so"
        )
        r["Stacktrace"] = "long\ntrace"
        r["ThreadStacktrace"] = "thread\neven longer\ntrace"
        crashdb.update_traces(crash_id, r, "good retrace!")
        r = crashdb.download(crash_id)

        # test fixed version
        self.assertIsNone(crashdb.get_fixed_version(crash_id))
        crashdb.close_duplicate(r, crash_id, self.get_uncommon_description_report())
        self.assertEqual(
            crashdb.duplicate_of(crash_id), self.get_uncommon_description_report()
        )
        self.assertEqual(crashdb.get_fixed_version(crash_id), "invalid")
        crashdb.close_duplicate(r, crash_id, None)
        self.assertIsNone(crashdb.duplicate_of(crash_id))
        self.assertIsNone(crashdb.get_fixed_version(crash_id))

    def test_download_robustness(self):
        """download() of uncommon description formats"""
        # only ProblemType/Architecture/DistroRelease in description
        r = self.crashdb.download(self.get_uncommon_description_report())
        self.assertEqual(r["ProblemType"], "Package")
        self.assertEqual(r["Architecture"], "amd64")
        self.assertTrue(r["DistroRelease"].startswith("Ubuntu "))

    def test_escalation(self):
        """Escalating bugs with more than 10 duplicates"""
        launchpad_instance = os.environ.get("APPORT_LAUNCHPAD_INSTANCE") or "qastaging"
        db = CrashDatabase(
            os.environ.get("LP_CREDENTIALS"),
            {
                "distro": "ubuntu",
                "launchpad_instance": launchpad_instance,
                "escalation_tag": "omgkittens",
                "escalation_subscription": "apport-hackers",
            },
        )

        count = 0
        p = db.launchpad.people[db.options["escalation_subscription"]].self_link
        # needs to have 13 consecutive valid bugs without dupes
        first_dup = 10070
        try:
            for b in range(first_dup, first_dup + 13):
                count += 1
                sys.stderr.write(f"{b} ")
                db.close_duplicate(Report(), b, self.get_segv_report())
                b = db.launchpad.bugs[self.get_segv_report()]
                has_escalation_tag = db.options["escalation_tag"] in b.tags
                has_escalation_subscription = any(
                    s.person_link == p for s in b.subscriptions
                )
                if count <= 10:
                    self.assertFalse(has_escalation_tag)
                    self.assertFalse(has_escalation_subscription)
                else:
                    self.assertTrue(has_escalation_tag)
                    self.assertTrue(has_escalation_subscription)
        finally:
            for b in range(first_dup, first_dup + count):
                sys.stderr.write(f"R{b} ")
                db.close_duplicate(Report(), b, None)
        sys.stderr.write("\n")

    def test_marking_python_task_mangle(self):
        """Test source package task fixup for marking interpreter
        scrashes."""
        self._mark_needs_dupcheck(self.get_python_report())
        unchecked_before = self.crashdb.get_dup_unchecked()
        self.assertIn(self.get_python_report(), unchecked_before)

        # add an upstream task, and remove the package name from the
        # package task; _mark_dup_checked is supposed to restore the
        # package name
        b = self.crashdb.launchpad.bugs[self.get_python_report()]
        if b.private:
            b.private = False
            b.lp_save()
        t = b.bug_tasks[0]
        t.target = self.crashdb.launchpad.distributions["ubuntu"]
        t.lp_save()
        b.addTask(target=self.crashdb.launchpad.projects["coreutils"])

        r = self.crashdb.download(self.get_python_report())
        self.crashdb._mark_dup_checked(self.get_python_report(), r)

        unchecked_after = self.crashdb.get_dup_unchecked()
        self.assertNotIn(self.get_python_report(), unchecked_after)
        self.assertEqual(
            unchecked_before, unchecked_after.union(set([self.get_python_report()]))
        )

        # upstream task should be unmodified
        b = self.crashdb.launchpad.bugs[self.get_python_report()]
        self.assertEqual(b.bug_tasks[0].bug_target_name, "coreutils")
        self.assertEqual(b.bug_tasks[0].status, "New")
        self.assertEqual(b.bug_tasks[0].importance, "Undecided")

        # package-less distro task should have package name fixed
        self.assertEqual(b.bug_tasks[1].bug_target_name, "coreutils (Ubuntu)")
        self.assertEqual(b.bug_tasks[1].status, "New")
        self.assertEqual(b.bug_tasks[1].importance, "Medium")

        # should not confuse get_fixed_version()
        self.assertIsNone(self.crashdb.get_fixed_version(self.get_python_report()))

    def _generate_sigsegv_report(self, signal: int = 11) -> Report:
        """Create a test executable which will die with a SIGSEGV, generate
        a core dump for it, create a problem report with those two
        arguments (ExecutablePath and CoreDump) and call add_gdb_info().

        Return the apport.report.Report.
        """
        workdir = None
        orig_cwd = os.getcwd()
        pr = Report()
        try:
            workdir = tempfile.mkdtemp()
            atexit.register(shutil.rmtree, workdir)
            os.chdir(workdir)

            # create a test executable
            with open("crash.c", "w", encoding="utf-8") as fd:
                fd.write(
                    """
int f(x) {
    int* p = 0; *p = x;
    return x+1;
}
int main() { return f(42); }
"""
                )
            assert subprocess.call(["gcc", "-g", "crash.c", "-o", "crash"]) == 0
            assert os.path.exists("crash")

            # call it through gdb and dump core
            subprocess.call(
                [
                    "gdb",
                    "--batch",
                    "--ex",
                    "run",
                    "--ex",
                    "generate-core-file core",
                    "./crash",
                ],
                stdout=subprocess.PIPE,
            )
            assert os.path.exists("core")
            subprocess.check_call(["sync"])
            assert (
                subprocess.call(["readelf", "-n", "core"], stdout=subprocess.PIPE) == 0
            )

            pr["ExecutablePath"] = os.path.join(workdir, "crash")
            pr["CoreDump"] = (os.path.join(workdir, "core"),)
            pr["Signal"] = str(signal)

            pr.add_gdb_info()
        finally:
            os.chdir(orig_cwd)

        pr.add_package_info(self.test_package)
        pr.add_os_info()
        pr.add_gdb_info()
        pr.add_user_info()
        self.assertEqual(pr.standard_title(), "crash crashed with SIGSEGV in f()")

        # add some binary gibberish which isn't UTF-8
        pr["ShortGibberish"] = ' "]\xb6"\n'
        pr["LongGibberish"] = "a\nb\nc\ne\nf\n\xff\xff\xff\n\f"

        return pr



"""Apport package hook for apport itself.

This adds /var/log/apport.log and the file listing in /var/crash to the report.
"""

# Copyright 2007 Canonical Ltd.
# Author: Martin Pitt <martin.pitt@ubuntu.com>

import glob

import apport.hookutils

APPORT_LOG = "/var/log/apport.log"


def add_info(report):
    """Add Apport logs to the problem report."""
    apport.hookutils.attach_file_if_exists(report, APPORT_LOG, "ApportLog")
    reports = glob.glob("/var/crash/*")
    if reports:
        report["CrashReports"] = apport.hookutils.command_output(
            ["stat", "-c", "%a:%u:%g:%s:%y:%x:%n"] + reports
        )



# Copyright (C) 2022 Canonical Ltd.
# Author: Benjamin Drung <benjamin.drung@canonical.com>
#
# This program is free software; you can redistribute it and/or modify it
# under the terms of the GNU General Public License as published by the
# Free Software Foundation; either version 2 of the License, or (at your
# option) any later version.  See http://www.gnu.org/copyleft/gpl.html for
# the full text of the license.

"""Unit tests for data/apport."""

import argparse
import datetime
import errno
import io
import os
import pathlib
import shutil
import signal
import subprocess
import sys
import tempfile
import time
import unittest
from unittest.mock import MagicMock

import apport.fileutils
import apport.user_group
from tests.helper import import_module_from_file, pidfd_open
from tests.paths import get_data_directory

apport_binary = import_module_from_file(get_data_directory() / "apport")


class TestApport(unittest.TestCase):
    # pylint: disable=too-many-public-methods
    """Unit tests for data/apport."""

    @classmethod
    def setUpClass(cls):
        cls.orig_report_dir = apport.fileutils.report_dir

    @classmethod
    def tearDownClass(cls):
        apport.fileutils.report_dir = cls.orig_report_dir

    def setUp(self):
        self.workdir = tempfile.mkdtemp()
        self.addCleanup(shutil.rmtree, self.workdir)
        apport.fileutils.report_dir = os.path.join(self.workdir, "crash")
        self.report_dir = pathlib.Path(apport.fileutils.report_dir)

    @unittest.mock.patch("subprocess.run")
    def test_check_kernel_crash(self, run_mock):
        """Test found kernel crash dump."""
        self.report_dir.mkdir()
        vmcore = self.report_dir / "vmcore"
        vmcore.touch()
        apport_binary.check_kernel_crash()
        run_mock.assert_called_once_with(
            ["/usr/share/apport/kernel_crashdump"], check=False
        )

    def test_check_lock_not_writable(self) -> None:
        """Test check_lock() with not writable lock file."""
        os.environ["APPORT_LOCK_FILE"] = "/non-existing/apport.lock"
        try:
            with self.assertRaisesRegex(SystemExit, "^1$"):
                apport_binary.check_lock()
        finally:
            del os.environ["APPORT_LOCK_FILE"]

    @unittest.mock.patch("fcntl.lockf")
    def test_check_lock_taken(self, lockf_mock: MagicMock) -> None:
        """Test check_lock() with lock file taken by other process."""
        # Since this is a unit test, let the mock raise an exception instead
        # of letting the fcntl.lockf call run into the timeout signal.
        lockf_mock.side_effect = TimeoutError()
        with tempfile.NamedTemporaryFile("w") as lockfile:
            os.environ["APPORT_LOCK_FILE"] = lockfile.name
            try:
                with self.assertRaisesRegex(SystemExit, "^1$"):
                    apport_binary.check_lock()
            finally:
                del os.environ["APPORT_LOCK_FILE"]

    def test_refine_core_ulimit_huge(self) -> None:
        """Test refine_core_ulimit() with huge limit."""
        options = apport_binary.parse_arguments(
            ["-p", str(os.getpid()), "-d", "1", "-c", str(pow(2, 64))]
        )
        self.assertEqual(apport_binary.refine_core_ulimit(options), -1)

    @unittest.mock.patch("os.isatty")
    def test_init_error_log_is_tty(self, isatty_mock: MagicMock) -> None:
        """Test init_error_log() doing nothing on a TTY."""
        isatty_mock.return_value = True
        stderr = sys.stderr
        apport_binary.init_error_log()
        # Check sys.stderr to be unchanged
        self.assertEqual(sys.stderr, stderr)
        isatty_mock.assert_called_once_with(2)

    def test_pidfd_getpid_process_gone(self) -> None:
        """Test pidfd_getpid() for a pidfd where the process is gone."""
        with subprocess.Popen(["sleep", "3600"]) as test_process:
            try:
                pidfd = os.pidfd_open(test_process.pid)
            finally:
                test_process.kill()
        try:
            pid = apport_binary.pidfd_getpid(pidfd)
            self.assertEqual(pid, -1)
        finally:
            os.close(pidfd)

    def test_pidfd_getpid_on_stdout(self) -> None:
        """Test pidfd_getpid() on stdout which is an invalid pidfd."""
        fd = sys.stdout.fileno()
        with self.assertRaisesRegex(OSError, f"Bad file descriptor: {fd}"):
            apport_binary.pidfd_getpid(fd)

    def test_proc_pid_not_exist(self) -> None:
        """Test ProcPid().exits() returning False."""
        with apport_binary.ProcPid(os.getpid()) as proc_pid:
            self.assertFalse(proc_pid.exists("nonexistent"))

    def test_proc_pid_has_same_pid(self) -> None:
        """Test ProcPid.has_same_pid() for same process."""
        pid = os.getpid()
        with apport_binary.ProcPid(pid) as proc_pid, pidfd_open(pid) as pidfd:
            self.assertTrue(proc_pid.has_same_pid(pidfd))

    def test_proc_pid_has_same_pid_false(self) -> None:
        """Test ProcPid.has_same_pid() for different processes."""
        pid = os.getpid()
        with apport_binary.ProcPid(pid) as proc_pid, pidfd_open(1) as pidfd:
            self.assertFalse(proc_pid.has_same_pid(pidfd))

    def test_receive_arguments_via_socket_import_error(self) -> None:
        """Test receive_arguments_via_socket() fail to import systemd."""
        with (
            self.assertLogs(level="ERROR") as error_logs,
            unittest.mock.patch("builtins.__import__") as import_mock,
            self.assertRaisesRegex(SystemExit, "^0$"),
        ):
            import_mock.side_effect = ModuleNotFoundError("No module named 'systemd'")
            apport_binary.receive_arguments_via_socket()

        import_mock.assert_called_once()
        self.assertRegex(
            error_logs.output[0], "apport-forward.socket.*systemd python module"
        )

    def test_receive_arguments_via_socket_invalid_socket(self):
        """Test receive_arguments_via_socket with invalid socket."""
        try:
            # pylint: disable=import-outside-toplevel
            from systemd.daemon import listen_fds
        except ImportError:
            self.skipTest("systemd Python module not available")
        assert listen_fds
        self.assertNotIn("LISTEN_FDS", os.environ)
        with self.assertRaisesRegex(SystemExit, "^1$"):
            apport_binary.receive_arguments_via_socket()

    @unittest.mock.patch.object(apport_binary, "check_lock", MagicMock())
    @unittest.mock.patch.object(
        apport_binary, "is_same_ns", MagicMock(return_value=False)
    )
    @unittest.mock.patch.object(apport_binary, "forward_crash_to_container")
    def test_forward_crash_to_container(self, forward_mock: MagicMock) -> None:
        """process_crash_from_kernel_with_proc_pid() to forward crash to container."""
        fake_pid = 67890
        yesterday = int(time.clock_gettime(time.CLOCK_BOOTTIME) * 100) - 86400
        options = apport_binary.parse_arguments(
            ["-p", "12345", "-d", "1", "-P", str(fake_pid)]
        )
        with tempfile.TemporaryDirectory() as tmpdir:
            stat = pathlib.Path(tmpdir) / "stat"
            stat.write_text(f"{fake_pid} (name) ?{' x' * 18} {yesterday} ...\n")
            status = pathlib.Path(tmpdir) / "status"
            status.write_text("Name:\tname\nUid:\t0\t0\t0\t0\nGid:\t0\t0\t0\t0\n")
            exe = pathlib.Path(tmpdir) / "exe"
            os.symlink(sys.executable, exe)
            root = pathlib.Path(tmpdir) / "root"
            container_exe = root / pathlib.Path(sys.executable).relative_to("/")
            container_exe.parent.mkdir(parents=True)
            container_exe.touch(mode=0o755)
            cmdline = pathlib.Path(tmpdir) / "cmdline"
            cmdline.write_text("name\0")
            with apport_binary.ProcPid(fake_pid, tmpdir) as proc_pid:
                exit_code = apport_binary.process_crash_from_kernel_with_proc_pid(
                    options, proc_pid
                )

        self.assertEqual(exit_code, 0)
        forward_mock.assert_called_once()

    @unittest.mock.patch.object(apport_binary, "init_error_log", MagicMock())
    @unittest.mock.patch.object(apport_binary, "start_apport")
    def test_main_start(self, start_mock):
        """Test calling apport with --start."""
        self.assertEqual(apport_binary.main(["--start"]), 0)
        start_mock.assert_called_once_with()

    @unittest.mock.patch.object(apport_binary, "init_error_log", MagicMock())
    @unittest.mock.patch.object(apport_binary, "stop_apport")
    def test_main_stop(self, stop_mock):
        """Test calling apport with --stop."""
        self.assertEqual(apport_binary.main(["--stop"]), 0)
        stop_mock.assert_called_once_with()

    def test_start(self):
        """Test starting Apport crash handler."""
        open_mock = unittest.mock.mock_open()
        with unittest.mock.patch("builtins.open", open_mock):
            apport_binary.start_apport()
        open_mock.assert_called_with(
            "/proc/sys/kernel/core_pipe_limit", "w", encoding="utf-8"
        )
        self.assertEqual(open_mock.call_count, 3)

    def test_consistency_checks_replaced_process(self):
        """Test consistency_checks() for a replaced crash process ID."""
        pid = os.getpid()
        options = apport_binary.parse_arguments(["-p", str(pid), "-d", "1"])
        now = int(time.clock_gettime(time.CLOCK_BOOTTIME) * 100)
        crash_user = apport.user_group.get_process_user_and_group()
        with apport_binary.ProcPid(pid) as proc_pid:
            self.assertFalse(
                apport_binary.consistency_checks(options, now, proc_pid, crash_user)
            )

    def test_process_crash_from_kernel_replaced_process(self) -> None:
        """Test process_crash_from_kernel() to abort if process ID has been reused.

        Instead of actual replacing the process with a new one with the same ID,
        just use a different pidfd.
        """
        pidfd = os.pidfd_open(os.getpid())
        with (
            self.assertLogs(level="ERROR") as info_logs,
            subprocess.Popen(["sleep", "3600"]) as test_process,
        ):
            try:
                args = [
                    "-p",
                    str(test_process.pid),
                    "-d",
                    "1",
                    "-P",
                    str(test_process.pid),
                    "-F",
                    str(pidfd),
                ]
                options = apport_binary.parse_arguments(args)
                exit_code = apport_binary.process_crash_from_kernel(options)
            finally:
                test_process.kill()

        self.assertEqual(exit_code, 1)
        self.assertIn(
            f"The process {test_process.pid} has already been replaced by a new"
            f" process with the same ID. Ignoring crash.",
            info_logs.output[0],
        )

    @unittest.mock.patch.object(apport_binary, "check_lock", MagicMock())
    def test_consistency_checks_before_forwarding(self) -> None:
        """Test that consistency checks are done before forwarding to container."""
        fake_pid = 67890
        future = int(time.clock_gettime(time.CLOCK_BOOTTIME) * 100) + 3600
        options = apport_binary.parse_arguments(
            ["-p", "12345", "-d", "1", "-P", str(fake_pid)]
        )
        with (
            self.assertLogs(level="ERROR") as info_logs,
            tempfile.TemporaryDirectory() as tmpdir,
        ):
            stat = pathlib.Path(tmpdir) / "stat"
            stat.write_text(f"{fake_pid} (name) ?{' x' * 18} {future} ...\n")
            status = pathlib.Path(tmpdir) / "status"
            status.write_text("Name:\tname\nUid:\t42\t42\t42\t42\nGid:\t7\t7\t7\t7\n")
            with apport_binary.ProcPid(fake_pid, tmpdir) as proc_pid:
                exit_code = apport_binary.process_crash_from_kernel_with_proc_pid(
                    options, proc_pid
                )

        self.assertEqual(exit_code, 0)
        self.assertIn(
            "process was replaced after Apport started, ignoring", info_logs.output[0]
        )

    def test_consistency_checks_mismatching_uid(self):
        """Test consistency_checks() for a mitmatching UID."""
        pid = os.getpid()
        crash_user = apport.user_group.get_process_user_and_group()
        options = apport_binary.parse_arguments(
            [
                "-p",
                str(pid),
                "-d",
                "1",
                "-u",
                str(crash_user.uid + 1),
                "-g",
                str(crash_user.gid + 1),
            ]
        )
        with apport_binary.ProcPid(pid) as proc_pid:
            self.assertFalse(
                apport_binary.consistency_checks(options, 1, proc_pid, crash_user)
            )

    def test_stop(self):
        """Test stopping Apport crash handler."""
        open_mock = unittest.mock.mock_open()
        with unittest.mock.patch("builtins.open", open_mock):
            apport_binary.stop_apport()
        open_mock.assert_called_with(
            "/proc/sys/kernel/core_pattern", "w", encoding="utf-8"
        )
        self.assertEqual(open_mock.call_count, 3)

    @unittest.mock.patch("os.setresgid", MagicMock())
    @unittest.mock.patch("os.setresuid", MagicMock())
    @unittest.mock.patch.object(apport_binary, "_run_with_output_limit_and_timeout")
    @unittest.mock.patch("os.path.exists")
    def test_is_closing_session(
        self, path_exist_mock: MagicMock, run_mock: MagicMock
    ) -> None:
        """Test is_closing_session()."""
        path_exist_mock.return_value = True
        run_mock.return_value = (b"(false,)\n", b"some stderr output\n")
        crash_user = apport.user_group.UserGroupID(1337, 1337)
        with tempfile.TemporaryDirectory() as tmpdir:
            env = pathlib.Path(tmpdir) / "environ"
            env.write_text("DBUS_SESSION_BUS_ADDRESS=unix:path=/run/user/1337/bus\0")
            with apport_binary.ProcPid(12345, tmpdir) as proc_pid:
                self.assertEqual(
                    apport_binary.is_closing_session(proc_pid, crash_user), True
                )
        path_exist_mock.assert_called_once_with("/run/user/1337/bus")
        run_mock.assert_called_once()

    def test_is_closing_session_no_environ(self) -> None:
        """Test is_closing_session() with no DBUS_SESSION_BUS_ADDRESS."""
        crash_user = apport.user_group.UserGroupID(1337, 1337)
        with tempfile.TemporaryDirectory() as tmpdir:
            env = pathlib.Path(tmpdir) / "environ"
            env.write_text("DISPLAY=:0\0")
            with apport_binary.ProcPid(12345, tmpdir) as proc_pid:
                self.assertEqual(
                    apport_binary.is_closing_session(proc_pid, crash_user), False
                )

    def test_is_closing_session_no_determine_socket(self) -> None:
        """Test is_closing_session() cannot determine D-Bus socket."""
        crash_user = apport.user_group.UserGroupID(42, 42)
        with tempfile.TemporaryDirectory() as tmpdir:
            env = pathlib.Path(tmpdir) / "environ"
            env.write_text("DBUS_SESSION_BUS_ADDRESS=unix:/run/user/42/bus\0")
            with apport_binary.ProcPid(12345, tmpdir) as proc_pid:
                self.assertEqual(
                    apport_binary.is_closing_session(proc_pid, crash_user), False
                )

    def test_is_closing_session_socket_not_exists(self) -> None:
        """Test is_closing_session() where D-Bus socket does not exist."""
        assert not os.path.exists("/run/user/1337/bus")
        crash_user = apport.user_group.UserGroupID(1337, 1337)
        with tempfile.TemporaryDirectory() as tmpdir:
            env = pathlib.Path(tmpdir) / "environ"
            env.write_text("DBUS_SESSION_BUS_ADDRESS=unix:path=/run/user/1337/bus\0")
            with apport_binary.ProcPid(12345, tmpdir) as proc_pid:
                self.assertEqual(
                    apport_binary.is_closing_session(proc_pid, crash_user), False
                )

    @unittest.mock.patch("os.setresgid", MagicMock())
    @unittest.mock.patch("os.setresuid", MagicMock())
    @unittest.mock.patch.object(apport_binary, "_run_with_output_limit_and_timeout")
    @unittest.mock.patch("os.path.exists")
    def test_is_closing_session_gdbus_failure(
        self, path_exist_mock: MagicMock, run_mock: MagicMock
    ) -> None:
        """Test is_closing_session() with OSError from gdbus."""
        path_exist_mock.return_value = True
        run_mock.side_effect = OSError(
            errno.ENOENT, "No such file or directory: 'gdbus'"
        )
        crash_user = apport.user_group.UserGroupID(1337, 1337)
        with tempfile.TemporaryDirectory() as tmpdir:
            env = pathlib.Path(tmpdir) / "environ"
            env.write_text("DBUS_SESSION_BUS_ADDRESS=unix:path=/run/user/1337/bus\0")
            with apport_binary.ProcPid(12345, tmpdir) as proc_pid:
                self.assertEqual(
                    apport_binary.is_closing_session(proc_pid, crash_user), False
                )
        path_exist_mock.assert_called_once_with("/run/user/1337/bus")
        run_mock.assert_called_once()

    @unittest.mock.patch.object(apport_binary, "check_lock", MagicMock())
    @unittest.mock.patch.object(apport_binary, "init_error_log", MagicMock())
    def test_missing_proc_pid(self) -> None:
        """Test /proc/<pid> is already gone."""
        fake_pid = 2147483647
        assert not os.path.exists(f"/proc/{fake_pid}")
        with self.assertLogs(level="ERROR") as error_logs:
            with tempfile.NamedTemporaryFile() as stdin:
                stdin_mock = io.FileIO(stdin.name)
                with unittest.mock.patch("sys.stdin", return_value=stdin_mock):
                    exit_code = apport_binary.main(["-p", str(fake_pid), "-d", "1"])
                    self.assertEqual(exit_code, 1)
        self.assertIn("/proc/2147483647 not found", error_logs.output[0])

    @unittest.mock.patch.object(apport_binary, "check_lock", MagicMock())
    @unittest.mock.patch.object(apport_binary, "consistency_checks", MagicMock())
    @unittest.mock.patch.object(apport_binary, "init_error_log", MagicMock())
    @unittest.mock.patch.object(
        apport_binary, "is_closing_session", MagicMock(return_value=False)
    )
    @unittest.mock.patch.object(apport_binary, "is_systemd_watchdog_restart")
    def test_main_ignore_watchdog_restart(
        self, is_systemd_watchdog_restart_mock: MagicMock
    ) -> None:
        """Test main() to ignore watchdog restarts.

        The underlying is_systemd_watchdog_restart() function is mocked,
        which should be tested in a separate unit test.
        """
        is_systemd_watchdog_restart_mock.return_value = True
        args = [
            "-p",
            str(os.getpid()),
            "-d",
            "1",
            "-s",
            str(int(signal.SIGABRT)),
            "-c",
            "-1",
        ]
        with self.assertLogs(level="ERROR") as error_logs:
            with tempfile.NamedTemporaryFile() as stdin:
                stdin_mock = io.TextIOWrapper(io.FileIO(stdin.name))
                with unittest.mock.patch("sys.stdin", stdin_mock):
                    exit_code = apport_binary.main(args)

        self.assertEqual(exit_code, 0)
        is_systemd_watchdog_restart_mock.assert_called_once()
        self.assertIn("Ignoring systemd watchdog restart", error_logs.output[0])

    @unittest.mock.patch.object(apport_binary, "init_error_log", MagicMock())
    def test_non_existing_systemd_coredump(self) -> None:
        """Test main() to print error if systemd-coredump cannot be found."""
        try:
            # pylint: disable-next=import-outside-toplevel
            import systemd.journal
        except ImportError as error:
            self.skipTest(f"{error.name} Python module not available")
        assert systemd.journal

        with self.assertLogs(level="ERROR") as error_logs:
            exit_code = apport_binary.main(["--from-systemd-coredump", "42-5846584-0"])

        self.assertEqual(exit_code, 1)
        self.assertIn(
            "No journal log for systemd unit"
            " systemd-coredump@42-5846584-0.service found.",
            error_logs.output[0],
        )

    def test_parse_arguments(self) -> None:
        """Test parse_arguments()"""
        args = [
            "-p42477",
            "-s4",
            "-c0",
            "-d1",
            "-P42477",
            "-u1000",
            "-g1000",
            "-F3",
            "--",
            "!usr!bin!divide-by-zero",
        ]
        options = apport_binary.parse_arguments(args)

        expected_options = argparse.Namespace(
            pid=42477,
            signal_number=4,
            core_ulimit=0,
            dump_mode=1,
            global_pid=42477,
            pidfd=3,
            uid=1000,
            gid=1000,
            executable_path="/usr/bin/divide-by-zero",
            systemd_coredump_instance=None,
            start=False,
            stop=False,
        )
        self.assertEqual(options, expected_options)

    def test_parse_arguments_kernel_without_pidfd(self) -> None:
        """Test parse_arguments() from kernel without pidfd (%F) support."""
        args = [
            "-p42477",
            "-s4",
            "-c0",
            "-d1",
            "-P42477",
            "-u1000",
            "-g1000",
            "-F",
            "--",
            "!usr!bin!divide-by-zero",
        ]
        options = apport_binary.parse_arguments(args)

        expected_options = argparse.Namespace(
            pid=42477,
            signal_number=4,
            core_ulimit=0,
            dump_mode=1,
            global_pid=42477,
            pidfd=None,
            uid=1000,
            gid=1000,
            executable_path="/usr/bin/divide-by-zero",
            systemd_coredump_instance=None,
            start=False,
            stop=False,
        )
        self.assertEqual(options, expected_options)

    def test_systemd_journal_import_error(self) -> None:
        """Test handling missing systemd.journal library correctly."""
        with (
            self.assertLogs(level="ERROR") as error_logs,
            unittest.mock.patch("builtins.__import__") as import_mock,
            self.assertRaisesRegex(SystemExit, "^1$"),
        ):
            import_mock.side_effect = ModuleNotFoundError(
                "No module named 'systemd.journal'"
            )
            apport_binary.process_crash_from_systemd_coredump("0-383264-0")

        import_mock.assert_called_once()
        self.assertRegex(error_logs.output[0], "Please install python3-systemd")

    @unittest.mock.patch.object(apport_binary, "process_crash")
    @unittest.mock.patch.object(apport_binary, "get_systemd_coredump")
    def test_reading_core_from_journal_log(
        self, get_systemd_coredump_mock: MagicMock, process_crash_mock: MagicMock
    ) -> None:
        """Test _user_can_read_coredump via process_crash_from_systemd_coredump

        The core dump can be provided via the journal log. In this case
        _user_can_read_coredump should behave correctly.
        """
        get_systemd_coredump_mock.return_value = {
            "COREDUMP": b"(\xb5/\xfd$\x0ca\x00\x00mocked core\nG\xfe\xe0\x10",
            "COREDUMP_CMDLINE": "python3",
            "COREDUMP_CWD": "/",
            "COREDUMP_ENVIRON": "SHELL=/bin/bash\n",
            "COREDUMP_EXE": sys.executable,
            "COREDUMP_GID": 0,
            "COREDUMP_PID": 123456789,
            "COREDUMP_PROC_MAPS": "mocked /proc/<pid>/maps",
            "COREDUMP_PROC_STATUS": "mocked /proc/<pid>/status",
            "COREDUMP_SIGNAL": 11,
            "COREDUMP_SIGNAL_NAME": "SIGSEGV",
            "COREDUMP_TIMESTAMP": datetime.datetime(
                2024, 2, 19, 12, 10, 42, tzinfo=datetime.timezone.utc
            ),
            "COREDUMP_UID": 0,
        }
        process_crash_mock.side_effect = lambda report, *_args: report

        report = apport_binary.process_crash_from_systemd_coredump("3-12345-7")

        get_systemd_coredump_mock.assert_called_once_with("3-12345-7")
        process_crash_mock.assert_called_once()
        self.assertEqual(report.pid, 123456789)
        self.assertEqual(
            dict(report),
            {
                "Date": "Mon Feb 19 12:10:42 2024",
                "CoreDump": b"(\xb5/\xfd$\x0ca\x00\x00mocked core\nG\xfe\xe0\x10",
                "ExecutablePath": sys.executable,
                "ExecutableTimestamp": str(int(os.stat(sys.executable).st_mtime)),
                "ProblemType": "Crash",
                "ProcCmdline": "python3",
                "ProcCwd": "/",
                "ProcEnviron": "SHELL=/bin/bash",
                "ProcMaps": "mocked /proc/<pid>/maps",
                "ProcStatus": "mocked /proc/<pid>/status",
                "Signal": "11",
                "SignalName": "SIGSEGV",
            },
        )



# Copyright (c) 2006 - 2011 Canonical Ltd.
# Authors: Robert Collins <robert@ubuntu.com>
#          Martin Pitt <martin.pitt@ubuntu.com>
#
# This program is free software; you can redistribute it and/or modify it
# under the terms of the GNU General Public License as published by the
# Free Software Foundation; either version 2 of the License, or (at your
# option) any later version.  See http://www.gnu.org/copyleft/gpl.html for
# the full text of the license.

"""Test apport_python_hook.py."""

# TODO: Address following pylint complaints
# pylint: disable=invalid-name

import atexit
import os
import re
import shutil
import stat
import subprocess
import tempfile
import textwrap
import unittest
import unittest.mock

import apport.fileutils
import apport.report
from tests.paths import local_test_environment


class T(unittest.TestCase):
    """Test apport_python_hook.py."""

    maxDiff = None

    @classmethod
    def setUpClass(cls):
        cls.env = os.environ | local_test_environment()
        cls.orig_report_dir = apport.fileutils.report_dir
        apport.fileutils.report_dir = tempfile.mkdtemp()
        cls.env["APPORT_REPORT_DIR"] = apport.fileutils.report_dir
        atexit.register(shutil.rmtree, apport.fileutils.report_dir)

    @classmethod
    def tearDownClass(cls):
        apport.fileutils.report_dir = cls.orig_report_dir

    def tearDown(self):
        for f in apport.fileutils.get_all_reports():
            os.unlink(f)

    def _test_crash(self, extracode="", scriptname=None, relpath=False):
        """Create a test crash."""
        # put the script into /var/tmp, since that isn't ignored in the
        # hook
        if scriptname:
            script = scriptname
            fd = os.open(scriptname, os.O_CREAT | os.O_WRONLY)
        else:
            (fd, script) = tempfile.mkstemp(dir="/var/tmp")
            self.addCleanup(os.unlink, script)

        os.write(
            fd,
            f"""\
#!/usr/bin/env {os.getenv('PYTHON', 'python3')}
import apport_python_hook
apport_python_hook.install()

def func(x):
    raise Exception(b'This should happen. \\xe2\\x99\\xa5'.decode('UTF-8'))

{extracode}
func(42)
""".encode(),
        )
        os.close(fd)
        os.chmod(script, 0o755)

        if relpath:
            binary = os.path.join(".", os.path.basename(script))
            orig_cwd = os.getcwd()
            os.chdir(os.path.dirname(script))
        else:
            binary = script
            orig_cwd = "."
        env = self.env.copy()
        env["PYTHONPATH"] = f"{env.get('PYTHONPATH', orig_cwd)}:/my/bogus/path"
        try:
            process = subprocess.run(
                [binary, "testarg1", "testarg2"],
                check=False,
                stderr=subprocess.PIPE,
                env=env,
                text=True,
            )
        finally:
            os.chdir(orig_cwd)
        self.assertEqual(
            process.returncode,
            1,
            "crashing test python program exits with failure code",
        )
        if not extracode:
            self.assertIn("This should happen.", process.stderr)

        return script

    def test_general(self):
        """General operation of the Python crash hook."""
        script = self._test_crash()

        # did we get a report?
        reports = apport.fileutils.get_new_reports()
        pr = None
        self.assertEqual(len(reports), 1, "crashed Python program produced a report")
        self.assertEqual(
            stat.S_IMODE(os.stat(reports[0]).st_mode),
            0o640,
            "report has correct permissions",
        )

        pr = apport.report.Report()
        with open(reports[0], "rb") as f:
            pr.load(f)

        # check report contents
        expected_keys = [
            "InterpreterPath",
            "PythonArgs",
            "Traceback",
            "ProblemType",
            "ProcEnviron",
            "ProcStatus",
            "ProcCmdline",
            "Date",
            "ExecutablePath",
            "ProcMaps",
            "UserGroups",
        ]
        self.assertTrue(
            set(expected_keys).issubset(set(pr.keys())), "report has necessary fields"
        )
        self.assertIn("bin/python", pr["InterpreterPath"])
        self.assertEqual(pr["ExecutablePath"], script)
        self.assertEqual(pr["ExecutableTimestamp"], str(int(os.stat(script).st_mtime)))
        self.assertEqual(pr["PythonArgs"], f"['{script}', 'testarg1', 'testarg2']")
        self.assertTrue(pr["Traceback"].startswith("Traceback"))
        self.assertIn(
            "func\n    raise Exception(b'This should happen.", pr["Traceback"]
        )

    def test_existing(self):
        """Python crash hook overwrites seen existing files."""
        script = self._test_crash()

        # did we get a report?
        reports = apport.fileutils.get_new_reports()
        self.assertEqual(len(reports), 1, "crashed Python program produced a report")
        self.assertEqual(
            stat.S_IMODE(os.stat(reports[0]).st_mode),
            0o640,
            "report has correct permissions",
        )

        # touch report -> "seen" case
        apport.fileutils.mark_report_seen(reports[0])

        reports = apport.fileutils.get_new_reports()
        self.assertEqual(len(reports), 0)

        script = self._test_crash(scriptname=script)
        reports = apport.fileutils.get_new_reports()
        self.assertEqual(len(reports), 1)

        # "unseen" case
        script = self._test_crash(scriptname=script)
        reports = apport.fileutils.get_new_reports()
        self.assertEqual(len(reports), 1)

    def test_symlink(self):
        """Python crash of a symlinked program resolves to target."""
        script = self._test_crash()

        # load report for this
        reports = apport.fileutils.get_new_reports()
        self.assertEqual(len(reports), 1, "crashed Python program produced a report")
        pr1 = apport.Report()
        with open(reports[0], "rb") as f:
            pr1.load(f)
        for f in apport.fileutils.get_all_reports():
            os.unlink(f)

        script_link = os.path.join(os.path.dirname(script), "script-link")
        os.symlink(os.path.basename(script), script_link)
        self.addCleanup(os.unlink, script_link)

        # run script through symlink name
        process = subprocess.run(
            [script_link], check=False, env=self.env, stderr=subprocess.PIPE, text=True
        )
        self.assertEqual(
            process.returncode,
            1,
            "crashing test python program exits with failure code",
        )
        self.assertIn("This should happen.", process.stderr)

        # get report for symlinked crash
        reports = apport.fileutils.get_new_reports()
        self.assertEqual(len(reports), 1, "crashed Python program produced a report")
        pr2 = apport.Report()
        with open(reports[0], "rb") as f:
            pr2.load(f)

        # check report contents
        self.assertIn("bin/python", pr2["InterpreterPath"])
        self.assertEqual(pr1["ExecutablePath"], script)
        self.assertEqual(pr2["ExecutablePath"], script)
        self.assertEqual(pr1.crash_signature(), pr2.crash_signature())

    def test_no_argv(self):
        """With zapped sys.argv."""
        script = self._test_crash("import sys\nsys.argv = None")

        # did we get a report?
        reports = apport.fileutils.get_new_reports()
        pr = None
        self.assertEqual(len(reports), 1, "crashed Python program produced a report")
        self.assertEqual(
            stat.S_IMODE(os.stat(reports[0]).st_mode),
            0o640,
            "report has correct permissions",
        )

        pr = apport.report.Report()
        with open(reports[0], "rb") as f:
            pr.load(f)

        # check report contents
        expected_keys = [
            "InterpreterPath",
            "Traceback",
            "ProblemType",
            "ProcEnviron",
            "ProcStatus",
            "ProcCmdline",
            "Date",
            "ExecutablePath",
            "ProcMaps",
            "UserGroups",
        ]
        self.assertTrue(
            set(expected_keys).issubset(set(pr.keys())), "report has necessary fields"
        )
        self.assertIn("bin/python", pr["InterpreterPath"])
        self.assertEqual(pr["ExecutablePath"], script)
        if "ExecutableTimestamp" in pr:
            self.assertEqual(
                pr["ExecutableTimestamp"],
                str(int(os.stat(pr["ExecutablePath"]).st_mtime)),
            )
        self.assertTrue(pr["Traceback"].startswith("Traceback"))

    def test_python_env(self):
        """Python environmental variables appear in report."""
        self._test_crash()

        # did we get a report?
        reports = apport.fileutils.get_new_reports()
        pr = None
        self.assertEqual(len(reports), 1, "crashed Python program produced a report")

        pr = apport.report.Report()
        with open(reports[0], "rb") as f:
            pr.load(f)

        # check report contents
        self.assertIn("PYTHONPATH", pr["ProcEnviron"])
        self.assertIn("/my/bogus/path", pr["ProcEnviron"])

    def _assert_no_reports(self):
        """Assert that there are no crash reports."""
        reports = apport.fileutils.get_new_reports()
        self.assertEqual(
            len(reports), 0, f"no crash reports present (cwd: {os.getcwd()})"
        )

    def test_deleted_working_directory(self):
        """Relative Python script from deleted working directory."""
        orig_cwd = os.getcwd()
        try:
            with tempfile.TemporaryDirectory(dir="/var/tmp") as tmpdir:
                deleted_dir = os.path.join(tmpdir, "gone")
                os.mkdir(deleted_dir)
                os.chdir(deleted_dir)
                os.rmdir(deleted_dir)

                with open(
                    os.path.join(tmpdir, "script.py"), "w", encoding="utf-8"
                ) as script:
                    script.write(
                        "import apport_python_hook\n"
                        "apport_python_hook.install()\n"
                        "raise ValueError()\n"
                    )

                env = os.environ.copy()
                env["PYTHONPATH"] = orig_cwd
                process = subprocess.run(
                    ["python3", "../script.py"],
                    check=False,
                    env=env,
                    stdout=subprocess.PIPE,
                    stderr=subprocess.PIPE,
                )
        finally:
            os.chdir(orig_cwd)

        # Workaround for pytest bug.
        # See https://github.com/pytest-dev/pytest-cov/issues/541
        stderr = re.sub(
            r"^(Error processing [A-Za-z0-9 /-]*"
            r"init_cov_core.pth:\n[\S\W]*\n)?"
            r"pytest-cov: Failed to setup subprocess coverage.*\n",
            "",
            process.stderr.decode(),
        )

        self.assertEqual(
            stderr,
            textwrap.dedent(
                """\
                Traceback (most recent call last):
                  File "../script.py", line 3, in <module>
                    raise ValueError()
                ValueError
                """
            ),
        )
        self.assertEqual(process.stdout.decode(), "")
        self.assertEqual(process.returncode, 1)
        self._assert_no_reports()

    def test_interactive(self):
        """Interactive Python sessions never generate a report."""
        orig_cwd = os.getcwd()
        try:
            for d in ("/tmp", "/usr/local", "/usr"):
                os.chdir(d)
                env = os.environ.copy()
                env["PYTHONPATH"] = orig_cwd
                python = subprocess.run(
                    ["python3"],
                    check=False,
                    env=env,
                    input=b"import apport_python_hook\n"
                    b"apport_python_hook.install()\n"
                    b"raise ValueError",
                    stdout=subprocess.PIPE,
                    stderr=subprocess.PIPE,
                )
                self.assertNotEqual(python.returncode, 0)
                self.assertEqual(python.stdout.decode(), "")
                self.assertIn("ValueError", python.stderr.decode())
                self._assert_no_reports()
        finally:
            os.chdir(orig_cwd)

    def test_ignoring(self):
        """The Python crash hook respects the ignore list."""
        # put the script into /var/tmp, since that isn't ignored in the
        # hook
        (fd, script) = tempfile.mkstemp(dir="/var/tmp")
        try:
            with (
                tempfile.NamedTemporaryFile() as i_file,
                unittest.mock.patch("apport.report._ignore_file", i_file.name),
            ):
                os.write(
                    fd,
                    textwrap.dedent(
                        f"""\
                        #!/usr/bin/env {os.getenv('PYTHON', 'python3')}
                        import apport_python_hook
                        apport_python_hook.install()

                        # Inject the mocked ignore file path.
                        import apport.report
                        apport.report._ignore_file = "{i_file.name}"

                        def func(x):
                            raise Exception('This should happen.')

                        func(42)
                        """
                    ).encode("ascii"),
                )
                os.close(fd)
                os.chmod(script, 0o755)

                # ignore
                r = apport.report.Report()
                r["ExecutablePath"] = script
                r.mark_ignore()
                r = None

                process = subprocess.run(
                    [script, "testarg1", "testarg2"],
                    check=False,
                    env=self.env,
                    stdout=subprocess.PIPE,
                    stderr=subprocess.PIPE,
                )
                err = process.stderr.decode()
                self.assertEqual(
                    process.returncode,
                    1,
                    "crashing test python program exits with failure code",
                )
                self.assertIn("Exception: This should happen.", err)

        finally:
            os.unlink(script)

        # did we get a report?
        reports = apport.fileutils.get_new_reports()
        self.assertEqual(len(reports), 0)

    def test_no_flooding(self):
        """Limit successive reports."""
        count = 0
        limit = 5
        try:
            while count < limit:
                self._test_crash(scriptname="/var/tmp/pytestcrash")
                reports = apport.fileutils.get_new_reports()
                if not reports:
                    break
                self.assertEqual(
                    len(reports), 1, "crashed Python program produced one report"
                )
                apport.fileutils.mark_report_seen(reports[0])
                count += 1
        finally:
            os.unlink("/var/tmp/pytestcrash")

        self.assertGreater(count, 1)
        self.assertLess(count, limit)

    def test_generic_os_error(self):
        """Raise OSError with errno and no known subclass."""
        self._test_crash(
            extracode=textwrap.dedent(
                """\
                def g():
                    raise OSError(99, 'something bad')

                g()
                """
            )
        )
        pr = self._load_report()
        # we expect it to append errno
        exe = pr["ExecutablePath"]
        self.assertEqual(pr.crash_signature(), f"{exe}:OSError(99):{exe}@11:g")

    def test_generic_os_error_no_errno(self):
        """Raise OSError without errno and no known subclass."""
        self._test_crash(
            extracode=textwrap.dedent(
                """\
                def g():
                    raise OSError('something bad')

                g()
                """
            )
        )
        pr = self._load_report()
        # we expect it to not stumble over the missing errno
        exe = pr["ExecutablePath"]
        self.assertEqual(pr.crash_signature(), f"{exe}:OSError:{exe}@11:g")

    def test_getcwd_error(self):
        """Raise FileNotFoundError on os.getcwd() call."""
        for relpath in (True, False):
            with self.subTest(relpath=relpath):
                self._test_crash(
                    extracode=textwrap.dedent(
                        """\
                        import os
                        import tempfile

                        tempdir = tempfile.mkdtemp()
                        os.chdir(tempdir)
                        os.rmdir(tempdir)
                        os.getcwd()
                        """
                    ),
                    relpath=relpath,
                )
                pr = self._load_report()
                self.assertIn(":FileNotFoundError:", pr.crash_signature())
                self.assertIn("os.getcwd()\nFileNotFoundError", pr["Traceback"])

    def test_subclassed_os_error(self):
        """Raise OSError with known subclass."""
        self._test_crash(
            extracode=textwrap.dedent(
                """\
                def g():
                    raise OSError(2, 'no such file /notexisting')

                g()
                """
            )
        )
        pr = self._load_report()
        # we expect it to not append errno, as it's already encoded
        # in the subclass
        exe = pr["ExecutablePath"]
        self.assertEqual(pr.crash_signature(), f"{exe}:FileNotFoundError:{exe}@11:g")

    def _load_report(self):
        """Ensure that there is exactly one crash report and load it."""
        reports = apport.fileutils.get_new_reports()
        self.assertEqual(len(reports), 1, "crashed Python program produced a report")
        pr = apport.Report()
        with open(reports[0], "rb") as f:
            pr.load(f)
        return pr



"""Enhanced Thread with support for return values and exception propagation."""

# Copyright (C) 2007 Canonical Ltd.
# Author: Martin Pitt <martin.pitt@ubuntu.com>
#
# This program is free software; you can redistribute it and/or modify it
# under the terms of the GNU General Public License as published by the
# Free Software Foundation; either version 2 of the License, or (at your
# option) any later version.  See http://www.gnu.org/copyleft/gpl.html for
# the full text of the license.

# pylint: disable=invalid-name
# pylint: enable=invalid-name

import sys
import threading


class REThread(threading.Thread):
    """Thread with return values and exception propagation.

    The thread is marked as daemon thread. The entire Python program exits
    when no alive non-daemon threads are left.
    """

    def __init__(self, group=None, target=None, name=None, args=(), kwargs=None):
        """Initialize Thread, identical to threading.Thread.__init__()."""
        if kwargs is None:
            kwargs = {}

        threading.Thread.__init__(self, group, target, name, args, kwargs, daemon=True)
        self.__target = target
        self.__args = args
        self.__kwargs = kwargs
        self._retval = None
        self._exception = None

    def run(self):
        """Run target function, identical to threading.Thread.run()."""
        if self.__target:
            try:
                self._retval = self.__target(*self.__args, **self.__kwargs)
            except BaseException:  # pylint: disable=broad-except
                if sys:  # pylint: disable=using-constant-test
                    self._exception = sys.exc_info()

    def return_value(self):
        """Return value from target function.

        This can only be called after the thread has finished, i. e. when
        is_alive() is False and did not terminate with an exception.
        """
        assert not self.is_alive()
        assert not self._exception
        return self._retval

    def exc_info(self):
        """Return (type, value, traceback) of the exception caught in run()."""
        return self._exception

    def exc_raise(self):
        """Raise the exception caught in the thread.

        Do nothing if no exception was caught.
        """
        if self._exception:
            raise self._exception[1].with_traceback(self._exception[2])



"""Convenience functions for use in package hooks."""

# Copyright (C) 2008 - 2012 Canonical Ltd.
# Authors:
#   Matt Zimmerman <mdz@canonical.com>
#   Brian Murray <brian@ubuntu.com>
#   Martin Pitt <martin.pitt@ubuntu.com>
#
# This program is free software; you can redistribute it and/or modify it
# under the terms of the GNU General Public License as published by the
# Free Software Foundation; either version 2 of the License, or (at your
# option) any later version.  See http://www.gnu.org/copyleft/gpl.html for
# the full text of the license.

# pylint: disable=too-many-lines
# TODO: Address following pylint complaints
# pylint: disable=invalid-name

import base64
import datetime
import glob
import os
import re
import select
import shutil
import stat
import subprocess
import sys
import tempfile
import warnings

import apport.fileutils
from apport.packaging_impl import impl as packaging
from problem_report import ProblemReport

_invalid_key_chars_re = re.compile(r"[^0-9a-zA-Z_.-]")
_AGENT = None


def path_to_key(path):
    """Generate a valid report key name from a file path.

    This will replace invalid punctuation symbols with valid ones.
    """
    if isinstance(path, bytes):
        path = path.decode("UTF-8")
    return _invalid_key_chars_re.sub(".", path.replace(" ", "_"))


def attach_file_if_exists(report, path, key=None, overwrite=True, force_unicode=False):
    """Attach file contents if file exists.

    If key is not specified, the key name will be derived from the file
    name with path_to_key().

    If overwrite is True, an existing key will be updated. If it is False, a
    new key with '_' appended will be added instead.

    If the contents is valid UTF-8, or force_unicode is True, then the value
    will be a string, otherwise it will be bytes.
    """
    # Prevent directory traversal. Do it here too so it won't disclose if
    # a file exists or not.
    if "../" in path:
        return

    if not key:
        key = path_to_key(path)

    if os.path.exists(path):
        attach_file(report, path, key, overwrite, force_unicode)


def read_file(path, force_unicode=False):
    # TODO: Split into smaller functions/methods
    # pylint: disable=too-many-return-statements
    """Return the contents of the specified path.

    If the contents is valid UTF-8, or force_unicode is True, then the value
    will a string, otherwise it will be bytes.

    Upon error, this will deliver a text representation of the error,
    instead of failing.
    """
    try:
        # Prevent directory traversal
        if "../" in path:
            return "Error: invalid path."
        fd = os.open(path, os.O_NOFOLLOW | os.O_RDONLY | os.O_NONBLOCK)
        st = os.fstat(fd)
        # make sure there are no symlinks in the full path
        real_path = os.path.realpath(path)
        if st.st_ino != os.stat(real_path).st_ino or path != real_path:
            os.close(fd)
            return "Error: path contained symlinks."
        # make sure the file isn't a FIFO or symlink
        if stat.S_ISREG(st.st_mode):
            with os.fdopen(fd, "rb") as f:
                contents = f.read().strip()
        else:
            os.close(fd)
            return "Error: path was not a regular file."
        if force_unicode:
            return contents.decode("UTF-8", errors="replace")
        try:
            return contents.decode("UTF-8")
        except UnicodeDecodeError:
            return contents
    except OSError as error:
        return f"Error: {str(error)}"


def attach_file(report, path, key=None, overwrite=True, force_unicode=False):
    """Attach a file to the report.

    If key is not specified, the key name will be derived from the file
    name with path_to_key().

    If overwrite is True, an existing key will be updated. If it is False, a
    new key with '_' appended will be added instead.

    If the contents is valid UTF-8, or force_unicode is True, then the value
    will a string, otherwise it will be bytes.
    """
    if not key:
        key = path_to_key(path)

    # Do not clobber existing keys
    if not overwrite:
        while key in report:
            key += "_"
    report[key] = read_file(path, force_unicode=force_unicode)


def attach_conffiles(report, package, conffiles=None, ui=None):
    """Attach information about any modified or deleted conffiles.

    If conffiles is given, only this subset will be attached. If ui is given,
    ask whether the contents of the file may be added to the report; if this is
    denied, or there is no UI, just mark it as "modified" in the report.
    """
    modified = packaging.get_modified_conffiles(package)

    for path, contents in modified.items():
        if conffiles and path not in conffiles:
            continue

        key = f"modified.conffile.{path_to_key(path)}"
        if isinstance(contents, str) and (
            contents == "[deleted]" or contents.startswith("[inaccessible")
        ):
            report[key] = contents
            continue

        if ui:
            response = ui.yesno(
                f'It seems you have modified the contents of "{path}".  '
                f"Would you like to add the contents of it to your bug report?"
            )
            if response:
                report[key] = contents
            else:
                report[key] = "[modified]"
        else:
            report[key] = "[modified]"

        mtime = datetime.datetime.fromtimestamp(os.stat(path).st_mtime)
        report[f"mtime.conffile.{path_to_key(path)}"] = mtime.isoformat()


# pylint: disable-next=unused-argument
def attach_upstart_overrides(report, package):
    """Attach information about any Upstart override files."""
    warnings.warn(
        "apport.hookutils.attach_upstart_overrides() is obsolete."
        " Upstart is dead. Please drop this call.",
        PendingDeprecationWarning,
        stacklevel=2,
    )


# pylint: disable-next=unused-argument
def attach_upstart_logs(report, package):
    """Attach information about a package's session upstart logs."""
    warnings.warn(
        "apport.hookutils.attach_upstart_logs() is obsolete."
        " Upstart is dead. Please drop this call.",
        PendingDeprecationWarning,
        stacklevel=2,
    )


def attach_dmesg(report):
    """Attach information from the kernel ring buffer (dmesg).

    This will not overwrite already existing information.
    """
    if not report.get("CurrentDmesg", "").strip():
        report["CurrentDmesg"] = root_command_output(["dmesg"])


def attach_dmi(report):
    """Attach Desktop Management Interface (DMI) information to the report."""
    dmi_dir = "/sys/class/dmi/id"
    if os.path.isdir(dmi_dir):
        for f in os.listdir(dmi_dir):
            if f in {"subsystem", "uevent"}:
                continue
            p = os.path.realpath(f"{dmi_dir}/{f}")
            st = os.stat(p)
            # ignore the root-only ones, since they have serial numbers
            if not stat.S_ISREG(st.st_mode) or (st.st_mode & 4 == 0):
                continue

            try:
                value = read_file(p)
            except OSError:
                continue
            if value:
                report[f"dmi.{f.replace('_', '.')}"] = value

    # Use the hardware information to create a machine type.
    if "dmi.sys.vendor" in report and "dmi.product.name" in report:
        report["MachineType"] = (
            f"{report['dmi.sys.vendor']} {report['dmi.product.name']}"
        )


def attach_hardware(report):
    """Attach a standard set of hardware-related data to the report, including:

    - kernel dmesg (boot and current)
    - /proc/interrupts
    - /proc/cpuinfo
    - /proc/cmdline
    - /proc/modules
    - lspci -vvnn
    - lscpi -vt
    - lsusb
    - lsusb -v
    - lsusb -t
    - devices from udev
    - DMI information from /sys
    - prtconf (sparc)
    - pccardctl status/ident
    """
    attach_dmesg(report)

    attach_file(report, "/proc/interrupts", "ProcInterrupts")
    attach_file(report, "/proc/cpuinfo", "ProcCpuinfo")
    attach_file(report, "/proc/cmdline", "ProcKernelCmdLine")

    if os.path.exists("/sys/bus/pci"):
        report["Lspci"] = command_output(["lspci", "-vvnn"])
        report["Lspci-vt"] = command_output(["lspci", "-vt"])
    report["Lsusb"] = command_output(["lsusb"])
    report["Lsusb-v"] = command_output(["lsusb", "-v"])
    report["Lsusb-t"] = command_output(["lsusb", "-t"])
    report["ProcModules"] = command_output(["sort", "/proc/modules"])
    report["UdevDb"] = command_output(["udevadm", "info", "--export-db"])
    report["acpidump"] = root_command_output(["/usr/share/apport/dump_acpi_tables.py"])

    # anonymize partition labels
    labels = report["UdevDb"]
    labels = re.sub("ID_FS_LABEL=(.*)", "ID_FS_LABEL=<hidden>", labels)
    labels = re.sub("ID_FS_LABEL_ENC=(.*)", "ID_FS_LABEL_ENC=<hidden>", labels)
    labels = re.sub("by-label/(.*)", "by-label/<hidden>", labels)
    labels = re.sub("ID_FS_LABEL=(.*)", "ID_FS_LABEL=<hidden>", labels)
    labels = re.sub("ID_FS_LABEL_ENC=(.*)", "ID_FS_LABEL_ENC=<hidden>", labels)
    labels = re.sub("by-label/(.*)", "by-label/<hidden>", labels)
    report["UdevDb"] = labels

    attach_dmi(report)

    if command_available("prtconf"):
        report["Prtconf"] = command_output(["prtconf"])

    if command_available("pccardctl"):
        out = command_output(["pccardctl", "status"]).strip()
        if out:
            report["PccardctlStatus"] = out
        out = command_output(["pccardctl", "ident"]).strip()
        if out:
            report["PccardctlIdent"] = out


def attach_alsa_old(report):
    """(loosely based on http://www.alsa-project.org/alsa-info.sh)
    for systems where alsa-info is not installed
    (i e, *buntu 12.04 and earlier)
    """
    attach_file_if_exists(report, os.path.expanduser("~/.asoundrc"), "UserAsoundrc")
    attach_file_if_exists(
        report, os.path.expanduser("~/.asoundrc.asoundconf"), "UserAsoundrcAsoundconf"
    )
    attach_file_if_exists(report, "/etc/asound.conf")
    attach_file_if_exists(report, "/proc/asound/version", "AlsaVersion")
    attach_file(report, "/proc/cpuinfo", "ProcCpuinfo")

    report["AlsaDevices"] = command_output(["ls", "-l", "/dev/snd/"])
    report["AplayDevices"] = command_output(["aplay", "-l"])
    report["ArecordDevices"] = command_output(["arecord", "-l"])

    report["PciMultimedia"] = pci_devices(PCI_MULTIMEDIA)

    cards = []
    if os.path.exists("/proc/asound/cards"):
        with open("/proc/asound/cards", encoding="utf-8") as fd:
            for line in fd:
                if "]:" in line:
                    fields = line.lstrip().split()
                    cards.append(int(fields[0]))

    for card in cards:
        key = f"Card{card}.Amixer.info"
        report[key] = command_output(["amixer", "-c", str(card), "info"])
        key = f"Card{card}.Amixer.values"
        report[key] = command_output(["amixer", "-c", str(card)])

        for codecpath in glob.glob(f"/proc/asound/card{card}/codec*"):
            if os.path.isfile(codecpath):
                codec = os.path.basename(codecpath)
                key = f"Card{card}.Codecs.{path_to_key(codec)}"
                attach_file(report, codecpath, key=key)
            elif os.path.isdir(codecpath):
                codec = os.path.basename(codecpath)
                for name in os.listdir(codecpath):
                    path = os.path.join(codecpath, name)
                    key = (
                        f"Card{card}.Codecs"
                        f".{path_to_key(codec)}.{path_to_key(name)}"
                    )
                    attach_file(report, path, key)


def attach_alsa(report):
    """Attach ALSA subsystem information to the report."""
    if os.path.exists("/usr/sbin/alsa-info"):
        report["AlsaInfo"] = command_output(
            ["/usr/sbin/alsa-info", "--stdout", "--no-upload"]
        )
    elif os.path.exists("/usr/share/alsa-base/alsa-info.sh"):
        report["AlsaInfo"] = command_output(
            ["/usr/share/alsa-base/alsa-info.sh", "--stdout", "--no-upload"]
        )
    else:
        attach_alsa_old(report)

    report["AudioDevicesInUse"] = command_output(
        ["fuser", "-v"]
        + glob.glob("/dev/dsp*")
        + glob.glob("/dev/snd/*")
        + glob.glob("/dev/seq*")
    )

    if os.path.exists("/usr/bin/pacmd"):
        report["PulseList"] = command_output(["pacmd", "list"])

    if os.path.exists("/usr/bin/pa-info"):
        report["PaInfo"] = command_output(["/usr/bin/pa-info"])

    attach_dmi(report)
    attach_dmesg(report)


def command_available(command):
    """Is given command on the executable search path?"""
    if "PATH" not in os.environ:
        return False
    path = os.environ["PATH"]
    for element in path.split(os.pathsep):
        if not element:
            continue
        filename = os.path.join(element, command)
        if os.path.isfile(filename) and os.access(filename, os.X_OK):
            return True
    return False


def command_output(
    command,
    input=None,  # pylint: disable=redefined-builtin
    stderr=subprocess.STDOUT,
    keep_locale=False,
    decode_utf8=True,
):
    """Try to execute given command (list) and return its stdout.

    In case of failure, a textual error gets returned. This function forces
    LC_MESSAGES to C, to avoid translated output in bug reports.

    If decode_utf8 is True (default), the output will be converted to a string,
    otherwise left as bytes.
    """
    env = os.environ.copy()
    if not keep_locale:
        env["LC_MESSAGES"] = "C"
    try:
        sp = subprocess.run(
            command,
            check=False,
            input=input,
            stdout=subprocess.PIPE,
            stderr=stderr,
            env=env,
        )
    except OSError as error:
        return f"Error: {str(error)}"

    if sp.returncode == 0:
        res = sp.stdout.strip()
    else:
        res = (
            b"Error: command "
            + str(command).encode()
            + b" failed with exit code "
            + str(sp.returncode).encode()
            + b": "
            + sp.stdout
        )

    if decode_utf8:
        res = res.decode("UTF-8", errors="replace")
    return res


def _spawn_pkttyagent():
    global _AGENT  # pylint: disable=global-statement

    if _AGENT is not None:
        return
    if os.geteuid() == 0:
        return
    if not sys.stdin.isatty():
        return
    if not os.path.exists("/usr/bin/pkttyagent"):
        return

    try:
        (r, w) = os.pipe2(0)
    except OSError:
        return

    # closed by kill_pkttyagent(), pylint: disable=consider-using-with
    _AGENT = subprocess.Popen(
        ["pkttyagent", "--notify-fd", str(w), "--fallback"],
        close_fds=False,
        stdin=subprocess.PIPE,
        stdout=subprocess.PIPE,
    )

    os.close(w)

    with select.epoll() as epoll:
        while True:
            epoll.register(r, select.EPOLLIN)
            events = epoll.poll()
            for _, event_type in events:
                if event_type & select.EPOLLHUP:
                    os.close(r)
                    return


def kill_pkttyagent():
    """Kill pkttyagent (from PolicyKit) if it was started by Apport."""
    global _AGENT  # pylint: disable=global-statement

    if _AGENT is None:
        return

    _AGENT.terminate()
    _AGENT.wait()
    _AGENT = None


def _root_command_prefix():
    if os.getuid() == 0:
        return []
    if os.path.exists("/usr/bin/pkexec"):
        _spawn_pkttyagent()
        return ["pkexec"]
    # the package hook won't have everything it wanted but that's okay
    return []


def root_command_output(  # pylint: disable=redefined-builtin
    command, input=None, stderr=subprocess.STDOUT, decode_utf8=True
):
    """Try to execute given command (list) as root and return its stdout.

    This passes the command through pkexec, unless the caller is already root.

    In case of failure, a textual error gets returned.

    If decode_utf8 is True (default), the output will be converted to a string,
    otherwise left as bytes.
    """
    assert isinstance(command, list), "command must be a list"
    output = command_output(
        _root_command_prefix() + command,
        input,
        stderr,
        keep_locale=True,
        decode_utf8=decode_utf8,
    )
    return output


def attach_root_command_outputs(report, command_map):
    """Execute multiple commands as root and put their outputs into report.

    command_map is a keyname -> 'shell command' dictionary with the commands to
    run. They are all run through /bin/sh, so you need to take care of shell
    escaping yourself. To include stderr output of a command, end it with
    "2>&1".

    Just like root_command_output, this passes the command through pkexec,
    unless the caller is already root.

    This is preferable to using root_command_output() multiple times, as that
    will ask for the password every time.
    """
    wrapper_path = os.path.join(
        os.path.abspath(os.environ.get("APPORT_DATA_DIR", "/usr/share/apport")),
        "root_info_wrapper",
    )
    workdir = tempfile.mkdtemp()
    try:
        # create a shell script with all the commands
        script_path = os.path.join(workdir, ":script:")
        with open(script_path, "w", encoding="utf-8") as script:
            for keyname, command in command_map.items():
                assert hasattr(
                    command, "strip"
                ), "command must be a string (shell command)"
                # use "| cat" here, so that we can end commands with 2>&1
                # (otherwise it would have the wrong redirection order)
                script.write(f"{command} | cat > {os.path.join(workdir, keyname)}\n")

        # run script
        subprocess.run(
            _root_command_prefix() + [wrapper_path, script_path], check=False
        )

        # now read back the individual outputs
        for keyname in command_map:
            try:
                with open(os.path.join(workdir, keyname), "rb") as f:
                    buf = f.read().strip()
            except OSError:
                # this can happen if the user dismisses authorization in
                # _root_command_prefix
                continue
            # opportunistically convert to strings, like command_output()
            try:
                buf = buf.decode("UTF-8")
            except UnicodeDecodeError:
                pass
            if buf:
                report[keyname] = buf
            f.close()
    finally:
        shutil.rmtree(workdir)


def __filter_re_process(pattern, process):
    lines = ""
    # Get stdout while waiting for process to complete
    while process.poll() is None:
        for line in process.stdout:
            line = line.decode("UTF-8", errors="replace")
            if pattern.search(line):
                lines += line
    # Ensure all stdout is read after process completion
    for line in process.stdout:
        line = line.decode("UTF-8", errors="replace")
        if pattern.search(line):
            lines += line
    process.stdout.close()
    process.wait()
    if process.returncode == 0:
        return lines
    return ""


def recent_syslog(pattern, path=None, *, journald_only_system=True):
    """Extract recent system messages which match a regex.

    pattern should be a "re" object. By default, messages are read from
    the systemd journal, or /var/log/syslog; but when giving "path", messages
    are read from there instead.
    The journald_only_system parameter controls the scope of messages that are
    extracted when reading from the systemd journal. If set to True (the
    default), only messages from the system services are extracted. If set to
    False, all messages that the current user can see are extracted.
    """
    if path:
        command = ["tail", "-n", "10000", path]
    elif os.path.exists("/run/systemd/system"):
        command = ["journalctl", "--quiet", "-b", "-a"]
        if journald_only_system:
            command.append("--system")
    elif os.access("/var/log/syslog", os.R_OK):
        command = ["tail", "-n", "10000", "/var/log/syslog"]
    else:
        return ""
    with subprocess.Popen(command, stdout=subprocess.PIPE) as process:
        return __filter_re_process(pattern, process)


def xsession_errors(pattern=None):
    """Extract messages from ~/.xsession-errors.

    By default this parses out glib-style warnings, errors, criticals etc. and
    X window errors.  You can specify a "re" object as pattern to customize the
    filtering.

    Please note that you should avoid attaching the whole file to reports, as
    it can, and often does, contain sensitive and private data.
    """
    path = os.path.expanduser("~/.xsession-errors")
    if not os.path.exists(path) or not os.access(path, os.R_OK):
        return ""

    if not pattern:
        pattern = re.compile(
            r"^(\(.*:\d+\): \w+-(WARNING|CRITICAL|ERROR))"
            r"|(Error: .*No Symbols named)"
            r"|([^ ]+\[\d+\]: ([A-Z]+):)"
            r"|([^ ]-[A-Z]+ \*\*:)"
            r"|(received an X Window System error)"
            r"|(^The error was \')"
            r"|(^  \(Details: serial \d+ error_code)"
        )

    lines = ""
    with open(path, "rb") as f:
        for line in f:
            line = line.decode("UTF-8", errors="replace")
            if pattern.search(line):
                lines += line
    return lines


PCI_MASS_STORAGE = 0x01
PCI_NETWORK = 0x02
PCI_DISPLAY = 0x03
PCI_MULTIMEDIA = 0x04
PCI_MEMORY = 0x05
PCI_BRIDGE = 0x06
PCI_SIMPLE_COMMUNICATIONS = 0x07
PCI_BASE_SYSTEM_PERIPHERALS = 0x08
PCI_INPUT_DEVICES = 0x09
PCI_DOCKING_STATIONS = 0x0A
PCI_PROCESSORS = 0x0B
PCI_SERIAL_BUS = 0x0C


def pci_devices(*pci_classes):
    """Return a text dump of PCI devices attached to the system."""
    if not pci_classes:
        return command_output(["lspci", "-vvnn"])

    result = ""
    output = command_output(["lspci", "-vvmmnn"])
    for paragraph in output.split("\n\n"):
        pci_class = None
        slot = None

        for line in paragraph.split("\n"):
            try:
                key, value = line.split(":", 1)
            except ValueError:
                continue
            value = value.strip()
            key = key.strip()
            if key == "Class":
                n = int(value[-5:-1], 16)
                pci_class = (n & 0xFF00) >> 8
            elif key == "Slot":
                slot = value

        if pci_class and slot and pci_class in pci_classes:
            if result:
                result += "\n\n"
            result += command_output(["lspci", "-vvnns", slot]).strip()

    return result


def usb_devices():
    """Return a text dump of USB devices attached to the system."""
    # TODO: would be nice to be able to filter by interface class
    return command_output(["lsusb", "-v"])


def files_in_package(package, globpat=None):
    """Retrieve a list of files owned by package, optionally matching
    globpat."""
    files = packaging.get_files(package)
    if globpat:
        result = [f for f in files if glob.fnmatch.fnmatch(f, globpat)]
    else:
        result = files
    return result


def attach_gconf(report, package):  # pylint: disable=unused-argument
    """Obsolete."""
    # keeping a no-op function for some time to not break hooks


def attach_gsettings_schema(report, schema):
    """Attach user-modified gsettings keys of a schema."""
    cur_value = report.get("GsettingsChanges", "")

    defaults = {}  # schema -> key ->  value
    env = os.environ.copy()
    env["XDG_CONFIG_HOME"] = "/nonexisting"
    with subprocess.Popen(
        ["gsettings", "list-recursively", schema], env=env, stdout=subprocess.PIPE
    ) as gsettings:
        for line in gsettings.stdout:
            try:
                (schema_name, key, value) = line.split(None, 2)
                value = value.rstrip()
            except ValueError:
                continue  # invalid line
            defaults.setdefault(schema_name, {})[key] = value

    with subprocess.Popen(
        ["gsettings", "list-recursively", schema], stdout=subprocess.PIPE
    ) as gsettings:
        for line in gsettings.stdout:
            try:
                (schema_name, key, value) = line.split(None, 2)
                value = value.rstrip()
            except ValueError:
                continue  # invalid line

            if value != defaults.get(schema_name, {}).get(key, ""):
                if schema_name == b"org.gnome.shell" and key in {
                    b"command-history",
                    b"favorite-apps",
                }:
                    value = "redacted by apport"
                cur_value += f"{schema_name} {key} {value}\n"

    report["GsettingsChanges"] = cur_value


def attach_gsettings_package(report, package):
    """Attach user-modified gsettings keys of all schemas in a package."""
    for schema_file in files_in_package(
        package, "/usr/share/glib-2.0/schemas/*.gschema.xml"
    ):
        schema = os.path.basename(schema_file)[:-12]
        attach_gsettings_schema(report, schema)


def attach_journal_errors(report: ProblemReport, time_window: int = 10) -> None:
    """Attach journal warnings and errors.

    If the report contains a date, get the journal logs around that
    date (plus/minus the time_window in seconds). Otherwise attach the
    latest 1000 journal logs since the last boot.
    """
    if not os.path.exists("/run/systemd/system"):
        return

    crash_timestamp = report.get_timestamp()
    if crash_timestamp:
        before_crash = crash_timestamp - time_window
        after_crash = crash_timestamp + time_window
        args = [f"--since=@{before_crash}", f"--until=@{after_crash}"]
    else:
        args = ["-b", "--lines=1000"]
    report["JournalErrors"] = command_output(
        ["journalctl", "--priority=warning"] + args
    )


def attach_network(report):
    """Attach generic network-related information to report."""
    report["IpRoute"] = command_output(["ip", "route"])
    report["IpAddr"] = command_output(["ip", "addr"])
    report["PciNetwork"] = pci_devices(PCI_NETWORK)
    attach_file_if_exists(report, "/etc/network/interfaces", key="IfupdownConfig")

    for var in ("http_proxy", "ftp_proxy", "no_proxy"):
        if var in os.environ:
            report[var] = os.environ[var]


def attach_wifi(report):
    """Attach wireless (WiFi) network information to report."""
    report["WifiSyslog"] = recent_syslog(
        re.compile(
            r"(NetworkManager|modem-manager|dhclient|kernel|wpa_supplicant)"
            r"(\[\d+\])?:"
        )
    )
    report["IwConfig"] = re.sub(
        "ESSID:(.*)",
        "ESSID:<hidden>",
        re.sub(
            "Encryption key:(.*)",
            "Encryption key: <hidden>",
            re.sub(
                "Access Point: (.*)",
                "Access Point: <hidden>",
                command_output(["iwconfig"]),
            ),
        ),
    )
    report["RfKill"] = command_output(["rfkill", "list"])
    if os.path.exists("/sbin/iw"):
        iw_output = command_output(["iw", "reg", "get"])
    else:
        iw_output = "N/A"
    report["CRDA"] = iw_output

    attach_file_if_exists(report, "/var/log/wpa_supplicant.log", key="WpaSupplicantLog")


def attach_printing(report):
    """Attach printing information to the report.

    Based on http://wiki.ubuntu.com/PrintingBugInfoScript.
    """
    attach_file_if_exists(report, "/etc/papersize", "Papersize")
    attach_file_if_exists(report, "/var/log/cups/error_log", "CupsErrorLog")
    report["Locale"] = command_output(["locale"])
    report["Lpstat"] = command_output(["lpstat", "-v"])

    ppds = glob.glob("/etc/cups/ppd/*.ppd")
    if ppds:
        nicknames = command_output(["fgrep", "-H", "*NickName"] + ppds)
        report["PpdFiles"] = re.sub(
            r'/etc/cups/ppd/(.*).ppd:\*NickName: *"(.*)"', r"\g<1>: \g<2>", nicknames
        )

    report["PrintingPackages"] = package_versions(
        "foo2zjs",
        "foomatic-db",
        "foomatic-db-engine",
        "foomatic-db-gutenprint",
        "foomatic-db-hpijs",
        "foomatic-filters",
        "foomatic-gui",
        "hpijs",
        "hplip",
        "m2300w",
        "min12xxw",
        "c2050",
        "hpoj",
        "pxljr",
        "pnm2ppa",
        "splix",
        "hp-ppd",
        "hpijs-ppds",
        "linuxprinting.org-ppds",
        "openprinting-ppds",
        "openprinting-ppds-extra",
        "ghostscript",
        "cups",
        "cups-driver-gutenprint",
        "foomatic-db-gutenprint",
        "ijsgutenprint",
        "cupsys-driver-gutenprint",
        "gimp-gutenprint",
        "gutenprint-doc",
        "gutenprint-locales",
        "system-config-printer-common",
        "kdeprint",
    )


def attach_mac_events(report, profiles=None):
    """Attach MAC information and events to the report."""
    # Allow specifying a string, or a list of strings
    if isinstance(profiles, str):
        profiles = [profiles]

    mac_regex = r"audit\(|apparmor|selinux|security"
    mac_re = re.compile(mac_regex, re.IGNORECASE)
    aa_regex = 'apparmor="DENIED".+?profile=([^ ]+?)[ ]'
    aa_re = re.compile(aa_regex, re.IGNORECASE)

    if "KernLog" not in report:
        report["KernLog"] = "\n".join(
            re.findall(mac_re, root_command_output(["dmesg"]))
        )

    if "AuditLog" not in report and os.path.exists("/var/run/auditd.pid"):
        attach_root_command_outputs(
            report, {"AuditLog": f'egrep "{mac_regex}" /var/log/audit/audit.log'}
        )

    attach_file_if_exists(report, "/proc/version_signature", "ProcVersionSignature")
    attach_file(report, "/proc/cmdline", "ProcCmdline")

    for match in re.findall(
        aa_re, report.get("KernLog", "") + report.get("AuditLog", "")
    ):
        if not profiles:
            report.add_tags(["apparmor"])
            break

        try:
            if match[0] == '"':
                profile = match[1:-1]
            else:
                profile = bytes.fromhex(match).decode("UTF-8", errors="replace")
        except (IndexError, ValueError):
            continue

        for search_profile in profiles:
            if re.match(f"^{search_profile}$", profile):
                report.add_tags(["apparmor"])
                break


def attach_related_packages(report, packages):
    """Attach version information for related packages.

    In the future, this might also run their hooks.
    """
    report["RelatedPackageVersions"] = package_versions(*packages)


def package_versions(*packages):
    """Return a text listing of package names and versions.

    Arguments may be package names or globs, e. g. "foo*"
    """
    if not packages:
        return ""
    versions = []
    for package_pattern in packages:
        if not package_pattern:
            continue

        matching_packages = packaging.package_name_glob(package_pattern)

        if not matching_packages:
            versions.append((package_pattern, "N/A"))

        for package in sorted(matching_packages):
            try:
                version = packaging.get_version(package)
            except ValueError:
                version = "N/A"
            if version is None:
                version = "N/A"
            versions.append((package, version))

    package_width = max(len(version[0]) for version in versions)
    return "\n".join([f"{p:<{package_width}s} {v}" for (p, v) in versions])


def _get_module_license(module):
    """Return the license for a given kernel module."""
    try:
        modinfo = subprocess.run(
            ["/sbin/modinfo", module],
            check=False,
            stdout=subprocess.PIPE,
            stderr=subprocess.PIPE,
        )
        if modinfo.returncode != 0:
            return "invalid"
    except OSError:
        return None
    for line in modinfo.stdout.decode("UTF-8").splitlines():
        fields = line.split(":", 1)
        if len(fields) < 2:
            continue
        if fields[0] == "license":
            return fields[1].strip()

    return None


def nonfree_kernel_modules(module_list="/proc/modules"):
    """Check loaded modules and return a list of those which are not free."""
    try:
        with open(module_list, encoding="utf-8") as f:
            mods = [line.split()[0] for line in f]
    except OSError:
        return []

    nonfree = []
    for m in mods:
        s = _get_module_license(m)
        if s and not ("GPL" in s or "BSD" in s or "MPL" in s or "MIT" in s):
            nonfree.append(m)

    return nonfree


def __drm_con_info(con):
    info = ""
    for f in os.listdir(con):
        path = os.path.join(con, f)
        if f == "uevent" or not os.path.isfile(path):
            continue
        with open(path, "rb") as con_info_file:
            val = con_info_file.read().strip()
        # format some well-known attributes specially
        if f == "modes":
            val = val.replace(b"\n", b" ")
        if f == "edid":
            val = base64.b64encode(val)
            f += "-base64"
        info += f"{f}: {val.decode('UTF-8', errors='replace')}\n"
    return info


def attach_drm_info(report):
    """Add information about DRM hardware.

    Collect information from /sys/class/drm/.
    """
    drm_dir = "/sys/class/drm"
    if not os.path.isdir(drm_dir):
        return
    for f in os.listdir(drm_dir):
        con = os.path.join(drm_dir, f)
        if os.path.exists(os.path.join(con, "enabled")):
            # DRM can set an arbitrary string for its connector paths.
            report[f"DRM.{path_to_key(f)}"] = __drm_con_info(con)


def in_session_of_problem(report):
    """Check if the problem happened in the currently running XDG session.

    This can be used to determine if e. g. ~/.xsession-errors is relevant and
    should be attached.

    Return None if this cannot be determined.
    """
    session_id = os.environ.get("XDG_SESSION_ID")
    if not session_id:
        # fall back to reading cgroup
        with open("/proc/self/cgroup", encoding="utf-8") as f:
            for line in f:
                line = line.strip()
                if (
                    "name=systemd:" in line
                    and line.endswith(".scope")
                    and "/session-" in line
                ):
                    session_id = line.split("/session-", 1)[1][:-6]
                    break
            else:
                return None

    try:
        report_time = report.get_timestamp()
    except AttributeError:
        return None
    if report_time is None:
        return None

    # determine session creation time
    try:
        session_start_time = os.stat(f"/run/systemd/sessions/{session_id}").st_mtime
    except OSError:
        return None

    return session_start_time <= report_time


def attach_default_grub(report, key=None):
    """Attach /etc/default/grub after filtering out password lines."""
    path = "/etc/default/grub"
    if not key:
        key = path_to_key(path)

    if os.path.exists(path):
        with open(path, "r", encoding="utf-8") as f:
            filtered = [
                (
                    line
                    if not line.startswith("password")
                    else "### PASSWORD LINE REMOVED ###"
                )
                for line in f.readlines()
            ]
            report[key] = "".join(filtered)


# backwards compatible API
shared_libraries = apport.fileutils.shared_libraries
links_with_shared_library = apport.fileutils.links_with_shared_library



"""Test dupdb-admin"""

# Copyright (C) 2022 Canonical Ltd.
# Author: Benjamin Drung <bdrung@ubuntu.com>
#
# This program is free software; you can redistribute it and/or modify it
# under the terms of the GNU General Public License as published by the
# Free Software Foundation; either version 2 of the License, or (at your
# option) any later version.  See http://www.gnu.org/copyleft/gpl.html for
# the full text of the license.

import os
import shutil
import subprocess
import tempfile
import unittest

from apport.crashdb_impl.memory import CrashDatabase
from tests.paths import local_test_environment


class TestDupdbAdmin(unittest.TestCase):
    # pylint: disable=missing-function-docstring
    """Test dupdb-admin"""

    def setUp(self):
        self.workdir = tempfile.mkdtemp()
        self.db_file = os.path.join(self.workdir, "apport_duplicates.db")
        self.env = os.environ | local_test_environment()

        self.crashes = CrashDatabase(
            None, {"sample_data": "1", "dupdb_url": f"file://{self.db_file}"}
        )
        self.crashes.init_duplicate_db(self.db_file)

    def tearDown(self):
        shutil.rmtree(self.workdir)

    def _call(
        self, args: list, expected_returncode: int = 0, expected_stdout: str | None = ""
    ) -> tuple[str, str]:
        cmd = ["dupdb-admin", "-f", self.db_file] + args
        process = subprocess.run(
            cmd,
            check=False,
            env=self.env,
            stdout=subprocess.PIPE,
            stderr=subprocess.PIPE,
            text=True,
        )
        self.assertEqual(process.returncode, expected_returncode, process.stderr)
        if process.returncode == 0:
            self.assertEqual(process.stderr, "")
        if expected_stdout is not None:
            self.assertEqual(process.stdout, expected_stdout)
        return (process.stdout, process.stderr)

    @staticmethod
    def _find_files_and_directories(base_dir: str) -> tuple[list[str], list[str]]:
        found_directories = []
        found_files = []
        for root, dirs, files in os.walk(base_dir):
            found_directories += [
                os.path.relpath(f"{root}/{d}", start=base_dir) for d in dirs
            ]
            found_files += [
                os.path.relpath(f"{root}/{f}", start=base_dir) for f in files
            ]
        return sorted(found_directories), sorted(found_files)

    def test_dump_empty_database(self):
        self._call(["dump"])

    def test_dump_database(self):
        self.assertIsNone(self.crashes.check_duplicate(0))
        self.assertIsNone(self.crashes.check_duplicate(2))
        self.crashes.duplicate_db_fixed(2, "42")
        stdout = self._call(["dump"], expected_stdout=None)[0]

        lines = stdout.rstrip().split("\n")
        self.assertEqual(len(lines), 2, lines)
        self.assertIn(
            "0: /bin/crash:11:foo_bar:d01:raise:<signal handler called>:__frob"
            " [open]",
            lines[0],
        )
        self.assertIn("2: /usr/bin/broken:11:h:g:f:e:d [fixed in: 42]", lines[1])

    def test_changeid(self):
        self.assertIsNone(self.crashes.check_duplicate(2))
        self._call(["changeid", "2", "1"])
        stdout = self._call(["dump"], expected_stdout=None)[0]
        self.assertIn("1: /usr/bin/broken:11:h:g:f:e:d [open]", stdout)

    def test_changeid_missing_argument(self):
        self.assertIsNone(self.crashes.check_duplicate(2))
        stderr = self._call(["changeid", "2"], expected_returncode=2)[1]
        self.assertIn("the following arguments are required: new_id", stderr)

    def test_missing_db_file(self):
        os.remove(self.db_file)
        stderr = self._call(["dump"], expected_returncode=1)[1]
        self.assertIn("file does not exist", stderr)

    def test_no_command(self):
        stderr = self._call([], expected_returncode=2)[1]
        self.assertIn("the following arguments are required: command", stderr)

    def test_publish(self):
        self.assertIsNone(self.crashes.check_duplicate(1))
        pub_path = f"{self.workdir}/www"
        self._call(["publish", pub_path])
        directories, files = self._find_files_and_directories(pub_path)
        self.assertEqual(directories, ["address", "sig"])
        self.assertEqual(files, ["sig/_bin_crash_11"])

    def test_publish_missing_argument(self):
        stderr = self._call(["publish"], expected_returncode=2)[1]
        self.assertIn("the following arguments are required: path", stderr)

    def test_removeid(self):
        self.assertIsNone(self.crashes.check_duplicate(1))
        self._call(["removeid", "1"])
        self._call(["dump"])

    def test_removeid_missing_argument(self):
        self.assertIsNone(self.crashes.check_duplicate(1))
        stderr = self._call(["removeid"], expected_returncode=2)[1]
        self.assertIn("the following arguments are required: id", stderr)

    def test_unknown_command(self):
        stderr = self._call(["nonexisting"], expected_returncode=2)[1]
        self.assertIn("invalid choice: 'nonexisting'", stderr)



"""Test functions in apport/packaging_impl/__init__.py."""

import unittest
import unittest.mock

from apport.packaging_impl import determine_packaging_implementation


class TestPackagingImpl(unittest.TestCase):
    # pylint: disable=missing-function-docstring
    """Test functions in apport/packaging_impl/__init__.py."""

    @unittest.mock.patch("platform.freedesktop_os_release")
    def test_determine_ubuntu(self, os_release_mock):
        os_release_mock.return_value = {
            "PRETTY_NAME": "Ubuntu 22.04.1 LTS",
            "NAME": "Ubuntu",
            "VERSION_ID": "22.04",
            "VERSION": "22.04.1 LTS (Jammy Jellyfish)",
            "VERSION_CODENAME": "jammy",
            "ID": "ubuntu",
            "ID_LIKE": "debian",
        }
        self.assertEqual(determine_packaging_implementation(), "apt_dpkg")
        os_release_mock.assert_called_once_with()

    @unittest.mock.patch("platform.freedesktop_os_release")
    def test_determine_debian_unstable(self, os_release_mock):
        os_release_mock.return_value = {
            "PRETTY_NAME": "Debian GNU/Linux bookworm/sid",
            "NAME": "Debian GNU/Linux",
            "ID": "debian",
        }
        self.assertEqual(determine_packaging_implementation(), "apt_dpkg")
        os_release_mock.assert_called_once_with()



# Copyright (C) 2012 Canonical Ltd.
# Author: Kyle Nitzsche <kyle.nitzsche@canonica.com>
#
# This program is free software; you can redistribute it and/or modify it
# under the terms of the GNU General Public License as published by the
# Free Software Foundation; either version 2 of the License, or (at your
# option) any later version.  See http://www.gnu.org/copyleft/gpl.html for
# the full text of the license.

"""System tests for bin/apport-valgrind."""

import os
import shutil
import subprocess
import tempfile
import unittest

import pytest

from tests.helper import skip_if_command_is_missing
from tests.paths import local_test_environment

with open("/proc/meminfo", encoding="utf-8") as f:
    for line in f.readlines():
        if line.startswith("MemTotal"):
            MEM_TOTAL_MiB = int(line.split()[1]) // 1024
            break


@skip_if_command_is_missing("valgrind")
class TestApportValgrind(unittest.TestCase):
    """System tests for bin/apport-valgrind."""

    @classmethod
    def setUpClass(cls):
        cls.env = os.environ | local_test_environment()

    def setUp(self):
        self.workdir = tempfile.mkdtemp()
        self.pwd = os.getcwd()

    def tearDown(self):
        shutil.rmtree(self.workdir)
        os.chdir(self.pwd)

    @pytest.mark.requires_internet
    @unittest.skipIf(MEM_TOTAL_MiB < 2000, f"{MEM_TOTAL_MiB} MiB is not enough memory")
    def test_sandbox_cache_options(self):
        """apport-valgrind creates a user specified sandbox and cache"""
        sandbox = os.path.join(self.workdir, "test-sandbox")
        cache = os.path.join(self.workdir, "test-cache")

        cmd = [
            "apport-valgrind",
            "--sandbox-dir",
            sandbox,
            "--cache",
            cache,
            "/usr/bin/true",
        ]
        subprocess.check_call(cmd, env=self.env)

        self.assertTrue(
            os.path.exists(sandbox),
            f"A sandbox directory {sandbox} was specified but was not created",
        )

        self.assertTrue(
            os.path.exists(cache),
            f"A cache directory {cache} was specified but was not created",
        )



"""Integration tests for the problem_report module."""

# TODO: Address following pylint complaints
# pylint: disable=invalid-name

import email
import gzip
import io
import os
import shutil
import tempfile
import textwrap
import time
import unittest

import problem_report

bin_data = b"ABABABABAB\0\0\0Z\x01\x02"


class T(unittest.TestCase):
    # pylint: disable=missing-class-docstring,missing-function-docstring

    def setUp(self):
        self.workdir = tempfile.mkdtemp()

    def tearDown(self):
        shutil.rmtree(self.workdir)

    def test_compressed_values(self):
        """Handle of CompressedValue values."""
        large_val = b"A" * 5000000

        pr = problem_report.ProblemReport()
        pr["Foo"] = problem_report.CompressedValue(b"FooFoo!")
        pr["Bin"] = problem_report.CompressedValue()
        pr["Bin"].set_value(bin_data)
        pr["Large"] = problem_report.CompressedValue(large_val)

        self.assertTrue(isinstance(pr["Foo"], problem_report.CompressedValue))
        self.assertTrue(isinstance(pr["Bin"], problem_report.CompressedValue))
        self.assertEqual(pr["Foo"].get_value(), b"FooFoo!")
        self.assertEqual(pr["Bin"].get_value(), bin_data)
        self.assertEqual(pr["Large"].get_value(), large_val)
        self.assertEqual(len(pr["Foo"]), 7)
        self.assertEqual(len(pr["Bin"]), len(bin_data))
        self.assertEqual(len(pr["Large"]), len(large_val))

        out = io.BytesIO()
        pr["Bin"].write(out)
        self.assertEqual(out.getvalue(), bin_data)
        out = io.BytesIO()
        pr["Large"].write(out)
        self.assertEqual(out.getvalue(), large_val)

        pr["Multiline"] = problem_report.CompressedValue(b"\1\1\1\n\2\2\n\3\3\3")
        self.assertEqual(pr["Multiline"].splitlines(), [b"\1\1\1", b"\2\2", b"\3\3\3"])

        # test writing of reports with CompressedValues
        out = io.BytesIO()
        pr.write(out)
        out.seek(0)
        pr = problem_report.ProblemReport()
        pr.load(out)
        self.assertEqual(pr["Foo"], "FooFoo!")
        self.assertEqual(pr["Bin"], bin_data)
        self.assertEqual(pr["Large"], large_val.decode("ASCII"))

    def test_write_append(self):
        """write() with appending to an existing file."""
        pr = problem_report.ProblemReport(date="now!")
        pr["Simple"] = "bar"
        pr["WhiteSpace"] = " foo   bar\nbaz\n  blip  "
        out = io.BytesIO()
        pr.write(out)

        pr.clear()
        pr["Extra"] = "appended"
        pr.write(out)

        self.assertEqual(
            out.getvalue().decode(),
            textwrap.dedent(
                f"""\
                ProblemType: Crash
                Date: now!
                Simple: bar
                WhiteSpace:
                  foo   bar
                 baz
                   blip{'  '}
                Extra: appended
                """
            ),
        )

        with tempfile.NamedTemporaryFile() as temp:
            temp.write(bin_data)
            temp.flush()

            pr = problem_report.ProblemReport(date="now!")
            pr["File"] = (temp.name,)
            out = io.BytesIO()
            pr.write(out)

        pr.clear()
        pr["Extra"] = "appended"
        pr.write(out)

        out.seek(0)
        pr = problem_report.ProblemReport()
        pr.load(out)

        self.assertEqual(pr["Date"], "now!")
        self.assertEqual(pr["File"], bin_data)
        self.assertEqual(pr["Extra"], "appended")

    def test_extract_keys(self):
        """extract_keys() with various binary elements."""
        # create a test report with binary elements
        large_val = b"A" * 5000000

        pr = problem_report.ProblemReport()
        pr["Txt"] = "some text"
        pr["MoreTxt"] = "some more text"
        pr["Foo"] = problem_report.CompressedValue(b"FooFoo!")
        pr["Uncompressed"] = bin_data
        pr["Bin"] = problem_report.CompressedValue()
        pr["Bin"].set_value(bin_data)
        pr["Large"] = problem_report.CompressedValue(large_val)
        pr["Multiline"] = problem_report.CompressedValue(b"\1\1\1\n\2\2\n\3\3\3")

        report = io.BytesIO()
        pr.write(report)
        report.seek(0)

        self.assertRaises(
            OSError,
            pr.extract_keys,
            report,
            "Bin",
            os.path.join(self.workdir, "nonexistent"),
        )
        # Test exception handling: Non-binary and nonexistent key
        tests = [
            (ValueError, "Txt"),
            (ValueError, ["Foo", "Txt"]),
            (KeyError, "Bar"),
            (KeyError, ["Foo", "Bar"]),
        ]
        for exc, keys_arg in tests:
            report.seek(0)
            self.assertRaises(exc, pr.extract_keys, report, keys_arg, self.workdir)

        # Check valid single elements
        tests = {
            "Foo": b"FooFoo!",
            "Uncompressed": bin_data,
            "Bin": bin_data,
            "Large": large_val,
            "Multiline": b"\1\1\1\n\2\2\n\3\3\3",
        }
        for key, expected in tests.items():
            report.seek(0)
            pr.extract_keys(report, key, self.workdir)
            with open(os.path.join(self.workdir, key), "rb") as f:
                self.assertEqual(f.read(), expected)
            # remove file for next pass
            os.remove(os.path.join(self.workdir, key))

        # Check element list
        report.seek(0)
        tests = {"Foo": b"FooFoo!", "Uncompressed": bin_data}
        pr.extract_keys(report, tests.keys(), self.workdir)
        for key, expected in tests.items():
            with open(os.path.join(self.workdir, key), "rb") as f:
                self.assertEqual(f.read(), expected)

    def test_write_file(self):
        """Write a report with binary file data."""
        with tempfile.NamedTemporaryFile() as temp:
            temp.write(bin_data)
            temp.flush()

            pr = problem_report.ProblemReport(date="now!")
            pr["File"] = (temp.name,)
            pr["Afile"] = (temp.name,)
            out = io.BytesIO()
            pr.write(out)

        self.assertIn("Afile: base64\n", out.getvalue().decode())
        self.assertIn("File: base64\n", out.getvalue().decode())
        report = problem_report.ProblemReport()
        out.seek(0)
        report.load(out)
        self.assertEqual(report["File"], bin_data)
        self.assertEqual(report["Afile"], bin_data)

        # force compression/encoding bool
        with tempfile.NamedTemporaryFile() as temp:
            temp.write(b"foo\0bar")
            temp.flush()
            pr = problem_report.ProblemReport(date="now!")
            pr["File"] = (temp.name, False)
            out = io.BytesIO()
            pr.write(out)

            self.assertEqual(
                out.getvalue().decode(),
                textwrap.dedent(
                    """\
                    ProblemType: Crash
                    Date: now!
                    File: foo\0bar
                    """
                ),
            )

            pr["File"] = (temp.name, True)
            out = io.BytesIO()
            pr.write(out)

            # check that written report is read correctly again
            report = problem_report.ProblemReport()
            out.seek(0)
            report.load(out, binary="compressed")
            self.assertIsInstance(report["File"], problem_report.CompressedValue)
            self.assertEqual(report["File"].get_value(), b"foo\0bar")

    def test_write_delayed_fileobj(self):
        """Write a report with file pointers and delayed data."""
        (fout, fin) = os.pipe()

        if os.fork() == 0:
            os.close(fout)
            time.sleep(0.3)
            os.write(fin, b"ab" * 512 * 1024)
            time.sleep(0.3)
            os.write(fin, b"hello")
            time.sleep(0.3)
            os.write(fin, b" world")
            os.close(fin)
            os._exit(0)  # pylint: disable=protected-access

        os.close(fin)

        pr = problem_report.ProblemReport(date="now!")
        out = io.BytesIO()
        with os.fdopen(fout, "rb") as f:
            pr["BinFile"] = (f,)
            pr.write(out)
        assert os.wait()[1] == 0

        out.seek(0)

        pr2 = problem_report.ProblemReport()
        pr2.load(out)
        self.assertEqual(pr2["BinFile"], "ab" * 512 * 1024 + "hello world")

    def test_big_file(self):
        """Write and re-decoding a big random file."""
        # create 1 MB random file
        with tempfile.NamedTemporaryFile() as temp:
            data = os.urandom(1048576)
            temp.write(data)
            temp.flush()

            # write it into problem report
            pr = problem_report.ProblemReport()
            pr["File"] = (temp.name,)
            pr["Before"] = "xtestx"
            pr["ZAfter"] = "ytesty"
            out = io.BytesIO()
            pr.write(out)

        # read it again
        out.seek(0)
        pr = problem_report.ProblemReport()
        pr.load(out)

        self.assertTrue(pr["File"] == data)
        self.assertEqual(pr["Before"], "xtestx")
        self.assertEqual(pr["ZAfter"], "ytesty")

        # write it again
        io2 = io.BytesIO()
        pr.write(io2)
        self.assertTrue(out.getvalue() == io2.getvalue())

        # check gzip compatibility
        out.seek(0)
        pr = problem_report.ProblemReport()
        pr.load(out, binary="compressed")
        self.assertEqual(pr["File"].get_value(), data)
        self.assertEqual(pr["File"].name, "File")

    def test_size_limit(self):
        """Write and a big random file with a size limit key."""
        # create 1 MB random file
        with tempfile.NamedTemporaryFile() as temp:
            data = os.urandom(1048576)
            temp.write(data)
            temp.flush()

            # write it into problem report
            pr = problem_report.ProblemReport()
            pr["FileSmallLimit"] = (temp.name, True, 100)
            pr["FileLimitMinus1"] = (temp.name, True, 1048575)
            pr["FileExactLimit"] = (temp.name, True, 1048576)
            pr["FileLimitPlus1"] = (temp.name, True, 1048577)
            pr["FileLimitNone"] = (temp.name, True, None)
            pr["Before"] = "xtestx"
            pr["ZAfter"] = "ytesty"
            out = io.BytesIO()
            pr.write(out)

        # read it again
        out.seek(0)
        pr = problem_report.ProblemReport()
        pr.load(out)

        self.assertNotIn("FileSmallLimit", pr)
        self.assertNotIn("FileLimitMinus1", pr)
        self.assertTrue(pr["FileExactLimit"] == data)
        self.assertTrue(pr["FileLimitPlus1"] == data)
        self.assertTrue(pr["FileLimitNone"] == data)
        self.assertEqual(pr["Before"], "xtestx")
        self.assertEqual(pr["ZAfter"], "ytesty")

    def test_add_to_existing(self):  # pylint: disable=too-many-statements
        """Add information to an existing report."""
        # original report
        pr = problem_report.ProblemReport()
        pr["old1"] = "11"
        pr["old2"] = "22"

        (fd, rep) = tempfile.mkstemp()
        os.close(fd)
        with open(rep, "wb") as f:
            pr.write(f)

        origstat = os.stat(rep)

        # create a new one and add it
        pr = problem_report.ProblemReport()
        pr.clear()
        pr["new1"] = "33"

        pr.add_to_existing(rep, keep_times=True)

        # check keep_times
        newstat = os.stat(rep)
        self.assertEqual(origstat.st_mode, newstat.st_mode)
        self.assertAlmostEqual(origstat.st_atime, newstat.st_atime, 1)
        self.assertAlmostEqual(origstat.st_mtime, newstat.st_mtime, 1)

        # check report contents
        newpr = problem_report.ProblemReport()
        with open(rep, "rb") as f:
            newpr.load(f)
        self.assertEqual(newpr["old1"], "11")
        self.assertEqual(newpr["old2"], "22")
        self.assertEqual(newpr["new1"], "33")

        # create a another new one and add it, but make sure mtime must be
        # different
        time.sleep(1)
        with open(rep, encoding="utf-8") as f:
            f.read()  # bump atime
        time.sleep(1)

        pr = problem_report.ProblemReport()
        pr.clear()
        pr["new2"] = "44"

        pr.add_to_existing(rep)

        # check that timestamps have been updates
        newstat = os.stat(rep)
        self.assertEqual(origstat.st_mode, newstat.st_mode)
        self.assertNotEqual(origstat.st_mtime, newstat.st_mtime)
        # skip atime check if filesystem is mounted noatime
        skip_atime = False
        directory = rep
        while len(directory) > 1:
            directory = os.path.split(directory)[0]
            if os.path.ismount(directory):
                with open("/proc/mounts", encoding="utf-8") as f:
                    for line in f:
                        mount, _, options = line.split(" ")[1:4]
                        if mount == directory and "noatime" in options.split(","):
                            skip_atime = True
                            break
                break
        if not skip_atime:
            self.assertNotEqual(origstat.st_atime, newstat.st_atime)

        # check report contents
        newpr = problem_report.ProblemReport()
        with open(rep, "rb") as f:
            newpr.load(f)
        self.assertEqual(newpr["old1"], "11")
        self.assertEqual(newpr["old2"], "22")
        self.assertEqual(newpr["new1"], "33")
        self.assertEqual(newpr["new2"], "44")

        os.unlink(rep)

    def test_write_mime_binary(self):
        """write_mime() for binary values and file references."""
        with tempfile.NamedTemporaryFile() as temp:
            with tempfile.NamedTemporaryFile() as tempgz:
                temp.write(bin_data)
                temp.flush()

                with gzip.GzipFile("File1", "w", fileobj=tempgz) as gz:
                    gz.write(bin_data)
                tempgz.flush()

                pr = problem_report.ProblemReport(date="now!")
                pr["Context"] = "Test suite"
                pr["File1"] = (temp.name,)
                pr["File1.gz"] = (tempgz.name,)
                pr["Value1"] = bin_data
                with open(tempgz.name, "rb") as f:
                    pr["Value1.gz"] = f.read()
                pr["ZValue"] = problem_report.CompressedValue(bin_data)
                out = io.BytesIO()
                pr.write_mime(out)
                out.seek(0)

        msg = email.message_from_binary_file(out)
        parts = list(msg.walk())
        self.assertEqual(len(parts), 7)

        # first part is the multipart container
        self.assertTrue(parts[0].is_multipart())

        # second part should be an inline text/plain attachments with all short
        # fields
        self.assertTrue(not parts[1].is_multipart())
        self.assertEqual(parts[1].get_content_type(), "text/plain")
        self.assertEqual(parts[1].get_content_charset(), "utf-8")
        self.assertIsNone(parts[1].get_filename())
        self.assertEqual(
            parts[1].get_payload(decode=True),
            b"ProblemType: Crash\nContext: Test suite\nDate: now!\n",
        )

        # third part should be the File1: file contents as gzip'ed attachment
        self.assertTrue(not parts[2].is_multipart())
        self.assertEqual(parts[2].get_content_type(), "application/gzip")
        self.assertEqual(parts[2].get_filename(), "File1.gz")
        self.assertEqual(self.decode_gzipped_message(parts[2]), bin_data)

        # fourth part should be the File1.gz: file contents as gzip'ed
        # attachment; write_mime() should not compress it again
        self.assertTrue(not parts[3].is_multipart())
        self.assertEqual(parts[3].get_content_type(), "application/gzip")
        self.assertEqual(parts[3].get_filename(), "File1.gz")
        self.assertEqual(self.decode_gzipped_message(parts[3]), bin_data)

        # fifth part should be the Value1: value as gzip'ed attachment
        self.assertTrue(not parts[4].is_multipart())
        self.assertEqual(parts[4].get_content_type(), "application/gzip")
        self.assertEqual(parts[4].get_filename(), "Value1.gz")
        self.assertEqual(self.decode_gzipped_message(parts[4]), bin_data)

        # sixth part should be the Value1: value as gzip'ed attachment;
        # write_mime should not compress it again
        self.assertTrue(not parts[5].is_multipart())
        self.assertEqual(parts[5].get_content_type(), "application/gzip")
        self.assertEqual(parts[5].get_filename(), "Value1.gz")
        self.assertEqual(self.decode_gzipped_message(parts[5]), bin_data)

        # seventh part should be the ZValue: value as gzip'ed attachment;
        # write_mime should not compress it again
        self.assertTrue(not parts[6].is_multipart())
        self.assertEqual(parts[6].get_content_type(), "application/gzip")
        self.assertEqual(parts[6].get_filename(), "ZValue.gz")
        self.assertEqual(self.decode_gzipped_message(parts[6]), bin_data)

    def test_write_mime_filter(self):
        """write_mime() with key filters."""
        pr = problem_report.ProblemReport(date="now!")
        pr["GoodText"] = "Hi"
        pr["BadText"] = "YouDontSeeMe"
        pr["GoodBin"] = bin_data
        pr["BadBin"] = "Y" + "\x05" * 10 + "-"
        out = io.BytesIO()
        pr.write_mime(out, skip_keys=["BadText", "BadBin"])
        out.seek(0)

        msg = email.message_from_binary_file(out)
        parts = list(msg.walk())
        self.assertEqual(len(parts), 3)

        # first part is the multipart container
        self.assertTrue(parts[0].is_multipart())

        # second part should be an inline text/plain attachments with all short
        # fields
        self.assertTrue(not parts[1].is_multipart())
        self.assertEqual(parts[1].get_content_type(), "text/plain")
        self.assertEqual(parts[1].get_content_charset(), "utf-8")
        self.assertIsNone(parts[1].get_filename())
        self.assertEqual(
            parts[1].get_payload(decode=True),
            textwrap.dedent(
                """\
                ProblemType: Crash
                Date: now!
                GoodText: Hi
                """
            ).encode(),
        )

        # third part should be the GoodBin: field as attachment
        self.assertTrue(not parts[2].is_multipart())
        self.assertEqual(self.decode_gzipped_message(parts[2]), bin_data)

    @staticmethod
    def decode_gzipped_message(message: email.message.Message) -> bytes:
        payload = message.get_payload(decode=True)
        assert isinstance(payload, bytes)
        with gzip.GzipFile(mode="rb", fileobj=io.BytesIO(payload)) as gzip_file:
            return gzip_file.read()
